from fastapi import FastAPI

from aa_rag import setting
from aa_rag.exceptions import handle_exception_error
from aa_rag.router import qa, solution, index, retrieve

app = FastAPI()
app.include_router(qa.router)
app.include_router(solution.router)
app.include_router(index.router)
app.include_router(retrieve.router)
app.add_exception_handler(Exception, handle_exception_error)


@app.get("/")
async def root():
    return {"message": "Hello World"}


@app.get("/default")
async def default():
    return setting.model_dump()


def startup():
    import uvicorn

    uvicorn.run(app, host=setting.server.host, port=setting.server.port)


if __name__ == "__main__":
    startup()



import logging
from os import PathLike
from pathlib import Path
from typing import Tuple, Optional
from urllib.parse import quote

import boto3
from botocore.exceptions import BotoCoreError, ClientError
from langchain_core.documents import Document
from pydantic import BaseModel, Field, HttpUrl, model_validator

from aa_rag import setting, utils


class OSSStoreInitParams(BaseModel):
    use_cache: bool = Field(
        default=True, examples=[True], description="Whether to use OSS cache."
    )
    update_cache: bool = Field(
        default=True, examples=[True], description="Whether to update OSS cache."
    )


class OSSResourceInfo(BaseModel):
    url: Optional[HttpUrl] = Field(default=None, description="A temp url from oss.")
    source_file_path: str = Field(..., description="The source file name.")
    hit_cache: bool = Field(..., description="Whether the url is from cache.")
    cache_file_path: str = Field(default=None, description="The cache file name.")
    version_id: str = Field(
        default=None,
        description="The version id of the file. If hit cache, the version id belong cache file, otherwise, source file.",
    )
    suffix: Optional[str] = Field(default=None, description="The suffix of the file.")

    @model_validator(mode="after")
    def check(self):
        if self.hit_cache:
            assert self.cache_file_path, (
                "The cache_file_path must be provided when hit_cache is True."
            )

        if self.suffix is None and self.url is not None:
            self.suffix = Path(self.url.path).suffix

        if not self.source_file_path.startswith(setting.oss.bucket):
            self.source_file_path = f"{setting.oss.bucket}/{self.source_file_path}"
        if not self.cache_file_path.startswith(setting.oss.cache_bucket):
            self.cache_file_path = f"{setting.oss.cache_bucket}/{self.cache_file_path}"

        return self


class OSSStore:
    from aa_rag.db.multimodal import StoreImageParams

    _oss_available: bool
    _oss_cache_available: bool

    def __init__(
        self,
        params: OSSStoreInitParams,
        oss_endpoint: str = setting.oss.endpoint,
        oss_bucket: str = setting.oss.bucket,
        oss_cache_bucket: str = setting.oss.cache_bucket,
        oss_access_key: str = setting.oss.access_key,
        oss_secret_key: str = setting.oss.secret_key.get_secret_value(),
    ):
        """
        Initialize the BaseParser with OSS settings and cache options.

        Args:
            params(OSSStoreInitParams)
            oss_endpoint (str): The endpoint URL for the OSS service.
            oss_bucket (str): The name of the main OSS bucket.
            oss_cache_bucket (str): The name of the cache OSS bucket.
            oss_access_key (str): The access key for the OSS service.
            oss_secret_key (str): The secret key for the OSS service.
        """
        self._oss_available, self._oss_cache_available = self._validate_oss(
            oss_endpoint, oss_bucket, oss_cache_bucket, oss_access_key, oss_secret_key
        )

        if self.oss_available:
            self.oss_client = boto3.client(
                "s3",
                endpoint_url=oss_endpoint,
                aws_access_key_id=oss_access_key,
                aws_secret_access_key=oss_secret_key,
                use_ssl=oss_endpoint.startswith("https://"),
                verify=oss_endpoint.startswith("https://"),
            )

        else:
            self.oss_client = None

        self.oss_bucket = oss_bucket
        self.oss_cache_bucket = oss_cache_bucket
        self.use_cache = params.use_cache
        self.update_cache = params.update_cache

    @property
    def oss_available(self):
        """
        Check if OSS is available.

        Returns:
            bool: True if OSS is available, False otherwise.
        """
        return self._oss_available

    @property
    def oss_cache_available(self):
        """
        Check if OSS cache is available.

        Returns:
            bool: True if OSS cache is available, False otherwise.
        """
        return self._oss_cache_available

    @staticmethod
    def _validate_oss(
        oss_endpoint: str,
        oss_bucket: str,
        oss_cache_bucket: str,
        oss_access_key: str,
        oss_secret_key: str,
    ) -> Tuple[bool, bool]:
        """
        Validate the OSS (Object Storage Service) connection and buckets.

        Args:
            oss_endpoint (str): The endpoint URL for the OSS service.
            oss_bucket (str): The name of the main OSS bucket.
            oss_cache_bucket (str): The name of the cache OSS bucket.
            oss_access_key (str): The access key for the OSS service.
            oss_secret_key (str): The secret key for the OSS service.

        Returns:
            Tuple[bool, bool]: A tuple of two boolean values. The first value indicates whether the OSS service is valid. The second value indicates whether the cache bucket is valid.
        """
        if not all([oss_access_key, oss_secret_key]):
            return False, False

        # Create S3 client
        try:
            oss_client = boto3.client(
                "s3",
                endpoint_url=oss_endpoint,
                aws_access_key_id=oss_access_key,
                aws_secret_access_key=oss_secret_key,
                use_ssl=oss_endpoint.startswith("https://"),
                verify=oss_endpoint.startswith("https://"),
            )
        except BotoCoreError as e:
            logging.warning(
                f"Failed to connect to OSS service: {str(e)}. No longer use OSS."
            )
            return False, False

        try:
            # Validate main bucket
            oss_client.head_bucket(Bucket=oss_bucket)
        except ClientError:
            logging.warning(
                f"Bucket not found: {oss_bucket} in oss service. No longer use OSS."
            )

        try:
            # Validate cache bucket
            oss_client.head_bucket(Bucket=oss_cache_bucket)
            return True, True
        except ClientError:
            logging.warning(
                f"Cache bucket not found: {oss_cache_bucket} in oss service. No longer use OSS cache."
            )
            return True, False

    def check_file_path(
        self, file_path: PathLike, **kwargs
    ) -> PathLike | OSSResourceInfo:
        """
        Check the file path and return the appropriate resource info.

        Args:
            file_path (PathLike): The file path to check.
            **kwargs: Additional keyword arguments.

        Returns:
            PathLike | OSSResourceInfo: The local file path or OSS resource info.
        """
        # find file path from local first
        if Path(file_path).exists():
            return file_path

        if self.oss_available:
            # check oss file exist
            try:
                oss_file_info = self.oss_client.head_object(
                    Bucket=setting.oss.bucket,
                    Key=str(file_path),
                    VersionId=kwargs.get("version_id", ""),
                )
            except ClientError:
                raise FileNotFoundError(
                    f"File not found: {file_path} in local and bucket: {setting.oss.bucket}"
                )

            md5_value = oss_file_info["ETag"].replace('"', "")
            cache_file_path = f"parsed_{md5_value}.md"
            if self.oss_cache_available and self.use_cache:
                # check oss cache file exist
                try:
                    cache_file_info = self.oss_client.head_object(
                        Bucket=setting.oss.cache_bucket,
                        Key=cache_file_path,
                        VersionId=kwargs.get("cache_version_id", ""),
                    )
                    target_bucket = self.oss_cache_bucket
                    target_file_path = cache_file_path
                    target_version_id = cache_file_info.get("VersionId")

                    hit_cache = True
                except ClientError:
                    target_bucket = self.oss_bucket
                    target_file_path = str(file_path)
                    target_version_id = oss_file_info.get("VersionId")

                    hit_cache = False
            else:
                target_bucket = self.oss_bucket
                target_file_path = file_path
                target_version_id = oss_file_info.get("VersionId")

                hit_cache = False

            # get temp url for file
            tmp_oss_url = self.oss_client.generate_presigned_url(
                "get_object",
                Params={
                    "Bucket": target_bucket,
                    "Key": target_file_path,
                    "VersionId": target_version_id,
                }
                if target_version_id
                else {
                    "Bucket": target_bucket,
                    "Key": target_file_path,
                },
            )
            return OSSResourceInfo(
                url=tmp_oss_url,
                source_file_path=str(file_path),
                cache_file_path=target_file_path if hit_cache else cache_file_path,
                hit_cache=hit_cache,
                version_id=target_version_id,
                suffix=None,
            )
        else:
            raise FileNotFoundError(f"File not found: {file_path} in local.")

    def store_image(self, params: StoreImageParams) -> Document:
        img_file_name, content_type, binary_data = (
            utils.convert_img_base64_to_file_info(params.image, params.img_desc)
        )
        img_file_path = f"image/{img_file_name}"

        try:
            oss_info = self.check_file_path(Path(img_file_path))
        except FileNotFoundError:
            self.oss_client.put_object(
                Bucket=self.oss_bucket,
                Key=str(Path(img_file_path)),
                Body=binary_data,
                ContentType=content_type,
                Metadata={
                    "description": quote(params.img_desc),
                },
            )
            oss_info = OSSResourceInfo(
                source_file_path=f"{self.oss_bucket}/{img_file_path}", hit_cache=False
            )

        assert isinstance(oss_info, OSSResourceInfo), (
            f"oss_info must be an instance of OSSResourceInfo, not {type(oss_info)}"
        )

        return Document(
            page_content=params.img_desc,
            metadata={
                "source": f"oss://{oss_info.source_file_path}",
                "suffix": Path(img_file_path).suffix,
            },
        )



import asyncio
import os
import tempfile
from abc import abstractmethod
from os import PathLike
from pathlib import Path
from typing import Iterable, List, Union, Dict, Literal

import aiohttp
import requests
from langchain_core.documents import Document

from aa_rag.oss import OSSStore, OSSStoreInitParams, OSSResourceInfo


class BaseParser(OSSStore):
    def __init__(self, use_cache: bool = True, update_cache: bool = True):
        """
        Initialize the BaseParser with OSS settings and cache options.

        Args:
            use_cache (bool, optional): Whether to use cache. Defaults to True.
            update_cache (bool, optional): Whether to update cache. Defaults to True.
        """

        super().__init__(
            OSSStoreInitParams(use_cache=use_cache, update_cache=update_cache)
        )

    @property
    def type(self):
        """
        Return the type of the parser. Must be implemented by subclasses.

        Returns:
            NotImplemented: This method should be overridden by subclasses.
        """
        return NotImplemented

    def parse(
        self,
        file_path: PathLike | Iterable[PathLike] = None,
        content: str | Iterable[str] = None,
        file_path_extra_kwargs: Dict = None,
        **kwargs,
    ) -> List[Document]:
        """
        Parse the provided file path(s) or content into a list of documents.

        Args:
            file_path (PathLike | Iterable[PathLike], optional): The file path(s) to parse. Defaults to None.
            content (str | Iterable[str], optional): The content to parse. Defaults to None.
            file_path_extra_kwargs (Dict, optional): Additional keyword arguments for each file path. Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Document]: A list of parsed documents.
        """
        assert file_path or content, "Either file_path or content must be provided."

        file_path_extra_kwargs = file_path_extra_kwargs or {}

        result: List[Document] = []

        if file_path:
            # file_path handling
            if isinstance(file_path, PathLike):
                file_path = [file_path]
            for _ in file_path:
                curr_uri: PathLike | OSSResourceInfo = self.check_file_path(
                    _, **file_path_extra_kwargs.get(_, {})
                )
                if isinstance(curr_uri, OSSResourceInfo):
                    with tempfile.NamedTemporaryFile(
                        delete=True, mode="wb", suffix=curr_uri.suffix
                    ) as temp_file:
                        response = requests.get(str(curr_uri.url), stream=True)
                        response.raise_for_status()

                        for chunk in response.iter_content(chunk_size=8192):
                            if chunk:
                                temp_file.write(chunk)

                        temp_file.flush()
                        os.fsync(temp_file.fileno())

                        temp_file.seek(0)

                        result.append(
                            self._parse_file(
                                Path(temp_file.name),
                                source="oss",
                                oss_resource_info=curr_uri,
                                **kwargs,
                            )
                        )

                    # update cache handling
                    if self.update_cache:
                        if not curr_uri.hit_cache:
                            self.oss_client.put_object(
                                Bucket=self.oss_cache_bucket,
                                Key=str(
                                    Path(curr_uri.cache_file_path).relative_to(
                                        self.oss_cache_bucket
                                    )
                                )
                                if curr_uri.cache_file_path.startswith(
                                    self.oss_cache_bucket
                                )
                                else curr_uri.cache_file_path,
                                Body=result[-1].page_content,
                            )

                elif isinstance(curr_uri, Path):
                    result.append(self._parse_file(curr_uri, source="local", **kwargs))

        if content:
            # content handling
            if isinstance(content, str):
                content = [content]
            for _ in content:
                result.append(self._parse_content(_, **kwargs))

        return result

    async def aparse(
        self,
        file_path: Union[PathLike, Iterable[PathLike]] = None,
        content: Union[str, Iterable[str]] = None,
        file_path_extra_kwargs: Dict = None,
        **kwargs,
    ) -> List[Document]:
        """
        Asynchronously parse the provided file path(s) or content into a list of documents.

        Args:
            file_path (Union[PathLike, Iterable[PathLike]], optional): The file path(s) to parse. Defaults to None.
            content (Union[str, Iterable[str]], optional): The content to parse. Defaults to None.
            file_path_extra_kwargs (Dict, optional): Additional keyword arguments for each file path. Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Document]: A list of parsed documents.
        """

        file_path_extra_kwargs = file_path_extra_kwargs or {}

        result: List[Document] = []

        if file_path:
            if isinstance(file_path, PathLike):
                file_path = [file_path]

            async def process_path(path):
                curr_uri = self.check_file_path(
                    path, **file_path_extra_kwargs.get(path, {})
                )
                if isinstance(curr_uri, OSSResourceInfo):
                    async with aiohttp.ClientSession() as session:
                        async with session.get(str(curr_uri.url)) as response:
                            url_content = await response.read()
                            with tempfile.NamedTemporaryFile(
                                delete=True, mode="wb", suffix=curr_uri.suffix
                            ) as temp_file:
                                temp_file.write(url_content)
                                temp_file.flush()

                                # update cache handling
                                if self.update_cache:
                                    if not curr_uri.hit_cache:
                                        self.oss_client.put_object(
                                            Bucket=self.oss_cache_bucket,
                                            Key=str(
                                                Path(
                                                    curr_uri.cache_file_path
                                                ).relative_to(self.oss_cache_bucket)
                                            )
                                            if curr_uri.cache_file_path.startswith(
                                                self.oss_cache_bucket
                                            )
                                            else curr_uri.cache_file_path,
                                            Body=result[-1].page_content,
                                        )

                                return self._parse_file(
                                    Path(temp_file.name),
                                    source="oss",
                                    oss_resource_info=curr_uri,
                                    **kwargs,
                                )

                elif isinstance(curr_uri, Path):
                    return self._parse_file(curr_uri, source="local", **kwargs)

            result += await asyncio.gather(*[process_path(p) for p in file_path])

        if content:
            if isinstance(content, str):
                content = [content]
            result += [self._parse_content(c, **kwargs) for c in content]

        return result

    @abstractmethod
    def _parse_file(
        self,
        file_path: PathLike,
        source: Literal["local", "oss"],
        oss_resource_info: OSSResourceInfo,
        **kwargs,
    ) -> Document:
        """
        Abstract method to parse a file. Must be implemented by subclasses.

        Args:
            file_path (PathLike): The file path to parse.
            source (Literal["local", "oss"]): The source of the file.
            oss_resource_info (OSSResourceInfo): The OSS resource info.
            **kwargs: Additional keyword arguments.

        Returns:
            Document: The parsed document.
        """
        return NotImplemented

    @abstractmethod
    def _parse_content(
        self, content: str, source: Literal["local", "oss"], **kwargs
    ) -> Document:
        """
        Abstract method to parse content. Must be implemented by subclasses.

        Args:
            content (str): The content to parse.
            source (Literal["local", "oss"]): The source of the content.
            **kwargs: Additional keyword arguments.

        Returns:
            Document: The parsed document.
        """
        return NotImplemented



import hashlib
import re
import uuid
from io import StringIO
from typing import List, Tuple

import pandas as pd
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from langchain_openai import OpenAIEmbeddings, ChatOpenAI

from aa_rag import setting
from aa_rag.db import LanceDBDataBase
from aa_rag.db.base import BaseVectorDataBase, BaseNoSQLDataBase
from aa_rag.db.milvus_ import MilvusDataBase
from aa_rag.db.mongo_ import MongoDBDataBase
from aa_rag.db.tinydb_ import TinyDBDataBase
from aa_rag.gtypes.enums import VectorDBType, NoSQLDBType, ParsingType
from aa_rag.gtypes.models.parse import ParserNeedItem
from aa_rag.parse.markitdown import MarkitDownParser
import base64
import mimetypes


def calculate_md5(input_string: str) -> str:
    """
    Calculate the MD5 hash of a string.

    Args:
        input_string (str): need to be calculated.

    Returns:
        str: MD5 hash of the input string.
    """
    md5_hash = hashlib.md5()
    md5_hash.update(input_string.encode("utf-8"))
    return md5_hash.hexdigest()


def get_embedding_model(
    model_name: str, return_dim: bool = False
) -> Embeddings | tuple[Embeddings, int]:
    """
    Get the embedding model based on the model name.
    Args:
        model_name (str): Model name.
        return_dim (bool): Return the embedding dimension if True.

    Returns:
        Embeddings: Embedding model instance.
        If return_dim is True, also returns the number of dimensions.

    """
    assert setting.openai.api_key, (
        "OpenAI API key is required for using OpenAI embeddings."
    )
    embeddings = OpenAIEmbeddings(
        model=model_name,
        dimensions=1536,
        api_key=setting.openai.api_key.get_secret_value(),
        base_url=setting.openai.base_url,
    )
    if return_dim:
        return embeddings, embeddings.dimensions or 1536
    else:
        return embeddings


def get_llm(model_name: str) -> BaseChatModel:
    assert setting.openai.api_key, (
        "OpenAI API key is required for using OpenAI embeddings."
    )
    model = ChatOpenAI(
        model=model_name,
        api_key=setting.openai.api_key,
        base_url=setting.openai.base_url,
    )

    return model


def get_vector_db(db_type: VectorDBType) -> BaseVectorDataBase | None:
    match db_type:
        case VectorDBType.LANCE:
            return LanceDBDataBase()
        case VectorDBType.MILVUS:
            return MilvusDataBase()
        case _:
            raise ValueError(f"Invalid db type: {db_type}")


def get_nosql_db(db_type: NoSQLDBType) -> BaseNoSQLDataBase | None:
    match db_type:
        case NoSQLDBType.TINYDB:
            return TinyDBDataBase()
        case NoSQLDBType.MONGODB:
            return MongoDBDataBase()
        case _:
            raise ValueError(f"Invalid db type: {db_type}")


def get_db(
    db_type: NoSQLDBType | VectorDBType,
) -> BaseNoSQLDataBase | BaseVectorDataBase | None:
    if isinstance(db_type, NoSQLDBType):
        return get_nosql_db(db_type)
    elif isinstance(db_type, VectorDBType):
        return get_vector_db(db_type)
    else:
        raise ValueError(f"Invalid db type: {db_type}")


def get_uuid():
    return str(uuid.uuid4()).replace("-", "")


async def parse_content(params: ParserNeedItem) -> List[Document]:
    if params.parsing_type == ParsingType.MARKITDOWN:
        parser = MarkitDownParser()
        source_data = await parser.aparse(
            **ParserNeedItem(**params.model_dump(exclude={"parsing_type"})).model_dump()
        )
    else:
        raise ValueError(f"Invalid parsing type: {params.parsing_type}")

    return source_data


def markdown_extract_csv_df(markdown_content):
    """将 Markdown 中的 CSV 内容转换为多个 DataFrame"""

    def extract_csv_sections(markdown_content):
        """从固定格式的 Markdown 中提取三个 CSV 部分"""
        # 定义正则表达式模式（注意 re.DOTALL 允许跨行匹配）
        pattern = re.compile(
            r"-----Entities-----\s*```csv\s*(.*?)\s*```"
            r".*?"
            r"-----Relationships-----\s*```csv\s*(.*?)\s*```"
            r".*?"
            r"-----Sources-----\s*```csv\s*(.*?)\s*```",
            re.DOTALL,
        )

        # 匹配并提取内容
        match = pattern.search(markdown_content)

        # 返回包含三个 CSV 内容的字典
        return (
            {
                "entities": match.group(1).strip() if match else "",
                "relationships": match.group(2).strip() if match else "",
                "sources": match.group(3).strip() if match else "",
            }
            if match
            else {}
        )

    # 提取原始 CSV 字符串
    raw_sections = extract_csv_sections(markdown_content)

    # 结果容器
    dfs = {}

    raw_sections.pop("sources")  # 移除 sources 部分

    # 处理每个 CSV 部分
    for section_name, csv_content in raw_sections.items():
        csv_content = csv_content.replace("\t", "")
        # 处理包含多行文本的特殊字段
        dfs[section_name] = pd.read_csv(
            StringIO(csv_content),
            escapechar="\\",  # 处理转义字符
            quotechar='"',  # 识别带逗号的字段
            skipinitialspace=True,
            on_bad_lines="warn",
        )

    return dfs["entities"], dfs["relationships"]


def convert_img_base64_to_file_info(
    base64_str: str, extra_calculate_value: str = None
) -> Tuple[str, str, bytes]:
    """
    将 Base64 字符串保存为图片文件名，文件名使用 Base64 数据的 MD5 哈希值

    Args:
        base64_str: 包含 Data URI 前缀（如 "data:image/png;base64,..."）或纯 Base64 的字符串
        extra_calculate_value: 额外的计算值用于计算MD5值

    Returns:
        返回文件名,文件类型，二进制的文件内容
    """
    # 初始化扩展名映射（处理常见 MIME 类型）
    ext_mapping = {
        "image/jpeg": ".jpg",
        "image/png": ".png",
        "image/gif": ".gif",
        "image/webp": ".webp",
    }

    # 分离 Data URI 和纯 Base64 数据
    if base64_str.startswith("data:"):
        header, data = base64_str.split(",", 1)
        mime_type = header.split(";")[0].split(":")[1]
    else:
        data = base64_str
        mime_type = None

    # 解码 Base64 数据
    binary_data = base64.b64decode(data)

    # 计算二进制数据的 MD5 哈希值
    value = (
        binary_data
        if extra_calculate_value is None
        else binary_data + extra_calculate_value.encode("utf-8")
    )
    md5_hash = hashlib.md5(value).hexdigest()

    # 确定文件扩展名
    if mime_type:
        # 优先从映射表获取扩展名
        ext = ext_mapping.get(mime_type)
        if not ext:
            # 使用 mimetypes 库猜测扩展名
            ext = mimetypes.guess_extension(mime_type) or ".bin"
    else:
        ext = ".bin"  # 无 MIME 类型时默认

    return f"{md5_hash}{ext}", mime_type, binary_data



from os import PathLike
from typing import Optional

from pydantic import Field

from aa_rag.gtypes.enums import ParsingType
from aa_rag.oss import OSSStoreInitParams


class ParserNeedItem(OSSStoreInitParams):
    file_path: Optional[PathLike] = Field(
        default=None,
        examples=[
            "user_manual/call_llm.md",
        ],
        description="Path to the file to be indexed. The file can from local file or OSS. Attention: The file_path and content cannot be both None.",
    )

    content: Optional[str] = Field(
        default=None,
        examples=[
            "# Call LLM\n\n## Introduction\n\nThis is a user manual for calling LLM.",
        ],
        description="The content to be indexed. Attention: The file_path and content cannot be both None",
    )

    parsing_type: ParsingType = Field(
        default=ParsingType.MARKITDOWN, description="The parsing type of the content."
    )

    # # Custom validator to ensure file_path and content are not both None
    # @model_validator(mode="after")
    # def check_file_path_or_content(self):
    #     # Check if both file_path and content are None
    #     if self.file_path is None and self.content is None:
    #         raise ValueError("Either file_path or content must be provided.")
    #
    #     return self    # # Custom validator to ensure file_path and content are not both None
    # @model_validator(mode="after")
    # def check_file_path_or_content(self):
    #     # Check if both file_path and content are None
    #     if self.file_path is None and self.content is None:
    #         raise ValueError("Either file_path or content must be provided.")
    #
    #     return self



from typing import List, cast, Union

import pandas as pd
from langchain.retrievers import EnsembleRetriever
from langchain_community.retrievers import BM25Retriever
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever
from langchain_milvus import Milvus
from langchain_text_splitters import RecursiveCharacterTextSplitter
from pydantic import BaseModel, Field

from aa_rag import setting, utils
from aa_rag.db.base import BaseDataBase, BaseVectorDataBase
from aa_rag.db.multimodal import StoreImageParams
from aa_rag.engine.base import BaseEngine
from aa_rag.gtypes.enums import EngineType, VectorDBType, DBMode, RetrieveType
from aa_rag.oss import OSSStore, OSSStoreInitParams

dfs_setting = setting.engine.simple_chunk


# 公共字段
class SimpleChunkInitParams(BaseModel):
    knowledge_name: str = Field(..., description="The name of the knowledge base.")
    identifier: str = Field(
        default="common", description="An optional identifier for the engine instance."
    )


# SimpleChunk 参数模型
class SimpleChunkIndexParams(StoreImageParams):
    source_data: Union[Document, List[Document]] = Field(
        ..., description="The source data to index."
    )

    retrieve_mode: DBMode = Field(
        default=setting.storage.mode, description="The mode of the index operation."
    )


class SimpleChunkRetrieveParams(BaseModel):
    query: str = Field(..., min_length=1)
    top_k: int = Field(
        dfs_setting.retrieve.k, description="The number of top results to return."
    )
    retrieve_mode: RetrieveType = Field(
        default=dfs_setting.retrieve.type, description="The retrieval method."
    )
    dense_weight: float = Field(
        default=dfs_setting.retrieve.weight.dense,
        description="The weight for dense retrieval.",
    )
    sparse_weight: float = Field(
        default=dfs_setting.retrieve.weight.sparse,
        description="The weight for sparse retrieval.",
    )


class SimpleChunkGenerateParams(BaseModel):
    pass


class SimpleChunk(
    BaseEngine[
        SimpleChunkIndexParams, SimpleChunkRetrieveParams, SimpleChunkGenerateParams
    ]
):
    @property
    def type(self):
        return EngineType.SimpleChunk

    def __init__(
        self,
        params: SimpleChunkInitParams,
        embedding_model: str = setting.embedding.model,
        db_type: VectorDBType = setting.storage.vector,
        **kwargs,
    ):
        """
        Initialize the SimpleChunk engine.

        Args:
            params (SimpleChunkInitParams): The initialization parameters.
            embedding_model (str, optional): The name of the embedding model to use. Defaults to setting.embedding.model.
            db_type (VectorDBType, optional): The type of vector database to use. Defaults to setting.storage.vector.
            **kwargs: Additional keyword arguments for the engine initialization.
        """
        # kwargs
        self.kwargs = kwargs

        self.embeddings, self.dimension = utils.get_embedding_model(
            embedding_model, return_dim=True
        )

        # parameters that make up the table name
        self.knowledge_name = params.knowledge_name
        self.embedding_model = embedding_model
        self.identifier = params.identifier

        # db
        self.db = utils.get_db(db_type)
        self.table_name = self._get_table(self.db)

    def retrieve(
        self,
        params: SimpleChunkRetrieveParams,
    ):
        """
        Retrieve documents based on the provided query and retrieval method.

        Args:
            params (SimpleChunkRetrieveParams): The retrieval parameters.
        Returns:
            List[Document]: A list of retrieved documents.
        """
        assert self.table_exist, f"Table {self.table_name} does not exist."

        query, top_k, retrieve_type = params.query, params.top_k, params.retrieve_mode

        match retrieve_type:
            case RetrieveType.DENSE:
                result = self._dense_retrieve(query, top_k)
            case RetrieveType.BM25:
                result = self._bm25_retrieve(query, top_k)
            case RetrieveType.HYBRID:
                result = self._hybrid_retrieve(
                    query,
                    top_k,
                    dense_weight=params.dense_weight,
                    sparse_weight=params.sparse_weight,
                )
            case _:
                raise ValueError(f"Invalid retrieve method: {retrieve_type}")

        return [doc.model_dump(include={"metadata", "page_content"}) for doc in result]

    def _get_table(self, db_obj: BaseDataBase) -> str:
        """
        Get table name in the vector database.

        Args:
            db_obj (BaseDataBase): The database object.

        Returns:
            str: The name of the table.
        """
        assert isinstance(db_obj, BaseVectorDataBase), (
            f"db_obj must be an instance of BaseVectorDataBase, not {type(db_obj)}"
        )

        table_name = f"{self.knowledge_name}_{self.type}_{self.embedding_model}"

        table_name = table_name.replace("-", "_")

        if table_name not in db_obj.table_list():
            self.table_exist = False
        else:
            self.table_exist = True

        return table_name

    def _create_table(self):
        vector_db: BaseVectorDataBase = cast(BaseVectorDataBase, self.db)

        if self.kwargs.get("schema"):
            schema = self.kwargs["schema"]
        else:
            match vector_db.db_type:
                case VectorDBType.LANCE:
                    import pyarrow as pa

                    schema = pa.schema(
                        [
                            pa.field("id", pa.utf8(), False),
                            pa.field(
                                "vector",
                                pa.list_(pa.float64(), self.dimension),
                                False,
                            ),
                            pa.field("text", pa.utf8(), False),
                            pa.field(
                                "metadata",
                                pa.struct(
                                    [
                                        pa.field("source", pa.utf8(), False),
                                    ]
                                ),
                                False,
                            ),
                        ]
                    )
                case VectorDBType.MILVUS:
                    from pymilvus import CollectionSchema, FieldSchema, DataType

                    id_field = FieldSchema(
                        name="id",
                        dtype=DataType.VARCHAR,
                        max_length=256,
                        is_primary=True,
                    )

                    vector_field = FieldSchema(
                        name="vector",
                        dtype=DataType.FLOAT_VECTOR,
                        dim=self.dimension,
                    )

                    text_field = FieldSchema(
                        name="text",
                        dtype=DataType.VARCHAR,
                        max_length=65535,
                    )

                    metadata_field = FieldSchema(
                        name="metadata",
                        dtype=DataType.JSON,
                    )

                    identifier_field = FieldSchema(
                        name="identifier",
                        dtype=DataType.JSON,
                    )

                    schema = CollectionSchema(
                        fields=[
                            id_field,
                            vector_field,
                            text_field,
                            metadata_field,
                            identifier_field,
                        ],
                    )
                case _:
                    raise ValueError(
                        f"Unsupported vector database type: {vector_db.db_type}"
                    )
        vector_db.create_table(self.table_name, schema=schema)

    def _dense_retrieve(
        self,
        query: str,
        top_k: int = dfs_setting.retrieve.k,
        only_return_retriever=False,
    ) -> BaseRetriever | List[Document]:
        """
        Perform a dense retrieval of documents based on the query.

        Args:
            query (str): The query string to search for.
            top_k (int, optional): The number of top results to return. Defaults to setting.retrieve.k.
            only_return_retriever (bool, optional): If True, only return the retriever object. Defaults to False.

        Returns:
            BaseRetriever | List[Document]: The retriever object if only_return_retriever is True, otherwise a list of retrieved documents.
        """
        vector_db, table_name = self.db, self.table_name
        assert isinstance(vector_db, BaseVectorDataBase), (
            f"db must be an instance of BaseVectorDataBase, not {type(vector_db)}"
        )

        # Get the appropriate retriever based on the vector database type
        match self.db.db_type:
            case VectorDBType.LANCE:
                from langchain_community.vectorstores import LanceDB

                dense_retriever = LanceDB(
                    connection=vector_db.connection,
                    table_name=table_name,
                    embedding=self.embeddings,
                )
            case VectorDBType.MILVUS:
                dense_retriever = Milvus(
                    embedding_function=self.embeddings,
                    collection_name=table_name,
                    connection_args={
                        **setting.storage.milvus.model_dump(
                            include={"uri", "user", "password"}
                        ),
                        "db_name": setting.storage.milvus.db_name,
                    },
                    primary_field="id",
                    metadata_field="metadata",
                )
            case _:
                raise ValueError(f"Unsupported vector database type: {self.db.db_type}")

        if only_return_retriever:
            dense_retriever = dense_retriever.as_retriever()
            dense_retriever.search_kwargs = {
                "expr": f'json_contains(identifier, "{self.identifier}")'
            }
            return dense_retriever

        # Perform the similarity search and return the results
        result: List[Document] = dense_retriever.similarity_search(
            query, k=top_k, expr=f'json_contains(identifier, "{self.identifier}")'
        )

        return result

    def _bm25_retrieve(
        self,
        query: str,
        top_k: int = dfs_setting.retrieve.k,
        only_return_retriever=False,
    ) -> BaseRetriever | List[Document]:
        """
        Perform a BM25 retrieval of documents based on the query.

        Args:
            query (str): The query string to search for.
            top_k (int, optional): The number of top results to return. Defaults to setting.retrieve.k.
            only_return_retriever (bool, optional): If True, only return the retriever object. Defaults to False.

        Returns:
            BaseRetriever | List[Document]: The retriever object if only_return_retriever is True, otherwise a list of retrieved documents.
        """
        vector_db, table_name = self.db, self.table_name
        assert isinstance(vector_db, BaseVectorDataBase), (
            f"db must be an instance of BaseVectorDataBase, not {type(vector_db)}"
        )

        # get retriever
        with vector_db.using(table_name) as table:
            all_doc = table.query(
                f"json_contains(identifier,'{self.identifier}')",
                limit=-1,
                output_fields=["id", "text", "metadata"],
            )  # get all documents
            all_doc_df = pd.DataFrame(all_doc)
            all_doc_s = (
                all_doc_df[["id", "text", "metadata"]]
                .apply(
                    lambda x: Document(
                        page_content=x["text"],
                        metadata={**x["metadata"], "id": x["id"]},
                    ),
                    axis=1,
                )
                .tolist()
            )
        sparse_retriever = BM25Retriever.from_documents(all_doc_s)
        sparse_retriever.k = top_k

        if only_return_retriever:
            return sparse_retriever

        # retrieve
        result: List[Document] = sparse_retriever.invoke(
            query, expr=f'json_contains(identifier, "{self.identifier}")'
        )

        return result

    def _hybrid_retrieve(
        self,
        query: str,
        top_k: int = dfs_setting.retrieve.k,
        dense_weight=dfs_setting.retrieve.weight.dense,
        sparse_weight=dfs_setting.retrieve.weight.sparse,
    ) -> List[Document]:
        """
        Perform a hybrid retrieval of documents based on the query.

        Args:
            query (str): The query string to search for.
            top_k (int, optional): The number of top results to return. Defaults to setting.retrieve.k.
            dense_weight (float, optional): The weight for dense retrieval. Defaults to setting.retrieve.weight.dense.
            sparse_weight (float, optional): The weight for sparse retrieval. Defaults to setting.retrieve.weight.sparse.

        Returns:
            List[Document]: A list of retrieved documents.
        """
        dense_retriever = self._dense_retrieve(query, top_k, only_return_retriever=True)
        sparse_retriever = self._bm25_retrieve(query, top_k, only_return_retriever=True)

        # combine the all retrievers
        ensemble_retriever = EnsembleRetriever(
            retrievers=[dense_retriever, sparse_retriever],
            weights=[dense_weight, sparse_weight],
        )
        return ensemble_retriever.invoke(
            query, expr=f'json_contains(identifier, "{self.identifier}")'
        )[:top_k]

    def index(
        self,
        params: SimpleChunkIndexParams,
        chunk_size: int = dfs_setting.index.chunk_size,
        chunk_overlap: int = dfs_setting.index.overlap_size,
    ):
        """
        Build index from source data and store to database.

        Args:
            params (SimpleChunkIndexParams): The index parameters.
            chunk_size (int, optional): The size of each chunk. Defaults to setting.index.chunk_size.
            chunk_overlap (int, optional): The overlap size between chunks. Defaults to setting.index.overlap_size.
        """
        if not self.table_exist:
            self._create_table()

        source_data, mode = params.source_data, params.retrieve_mode

        if isinstance(source_data, Document):
            source_docs = [source_data]
        else:
            source_docs = source_data

        # split the document into chunks
        splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )

        indexed_data = splitter.split_documents(source_docs)

        # handle image
        img_params: StoreImageParams = StoreImageParams(**params.model_dump())
        if img_params.image and img_params.img_desc:
            oss_store_params = OSSStoreInitParams(**img_params.model_dump())
            img_doc: Document = OSSStore(params=oss_store_params).store_image(
                img_params
            )
            indexed_data.append(img_doc)  # add image description to indexed_data

        # store index

        # detects whether the metadata has an id field. If not, it will be generated id based on page_content via md5 algorithm.
        id_s = [
            doc.metadata.get("id", utils.calculate_md5(doc.page_content))
            for doc in indexed_data
        ]

        text_vector_s = self.embeddings.embed_documents(
            [_.page_content for _ in indexed_data]
        )

        data = []

        for id_, vector, doc in zip(id_s, text_vector_s, indexed_data):
            data.append(
                {
                    "id": id_,
                    "vector": vector,
                    "text": doc.page_content,
                    "metadata": doc.metadata,
                    "identifier": [self.identifier],
                }
            )

        vector_db, table_name = self.db, self.table_name
        assert isinstance(vector_db, BaseVectorDataBase), (
            f"db must be an instance of BaseVectorDataBase, not {type(vector_db)}"
        )

        with vector_db.using(table_name) as table:
            match mode:
                case DBMode.INSERT:
                    table.add(data)

                case DBMode.UPSERT:
                    table.upsert(data)

                case DBMode.OVERWRITE:
                    table.overwrite(data)
                case _:
                    raise ValueError(f"Invalid mode: {mode}")

    def generate(self, params: SimpleChunkGenerateParams):
        return NotImplemented



from fastapi import status
from fastapi.responses import JSONResponse

from aa_rag.gtypes.models.base import BaseResponse


async def handle_exception_error(request, exc):
    """
    Handle universal exception error
    Args:
        request:
        exc:

    Returns:

    """
    return JSONResponse(
        status_code=status.HTTP_400_BAD_REQUEST,
        content=BaseResponse(
            code=status.HTTP_400_BAD_REQUEST,
            status="failed",
            message=f"{type(exc).__name__} Error",
            data=str(exc),
        ).model_dump(),
    )



from abc import abstractmethod, ABC
from typing import Any, List

from pandas import DataFrame

from aa_rag.gtypes.enums import VectorDBType, NoSQLDBType


def singleton(cls):
    instances = {}

    def wrapper(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return wrapper


class BaseDataBase:
    _db_type: Any
    _conn_obj: Any

    def __init__(self, **kwargs):
        self._conn_obj = self.connect(**kwargs)

    @property
    def connection(self):
        return self._conn_obj

    @property
    def db_type(self):
        return self._db_type

    @abstractmethod
    def connect(self, **kwargs):
        return NotImplemented

    @abstractmethod
    def table_list(self) -> List[str]:
        return NotImplemented

    @abstractmethod
    def create_table(self, table_name, schema, **kwargs):
        return NotImplemented

    @abstractmethod
    def drop_table(self, table_name):
        return NotImplemented

    @abstractmethod
    def close(self):
        return NotImplemented

    @abstractmethod
    def using(self, *args, **kwargs):
        """
        Set the table or collection to use. This is useful for chaining methods.

        Returns:
            self

        """
        return self

    def __enter__(self):
        return self

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        return NotImplemented

    def __call__(self, *args, **kwargs):
        return self.using(*args, **kwargs)


class BaseVectorDataBase(BaseDataBase, ABC):
    _db_type: VectorDBType

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def insert(self, data: list[dict] | DataFrame, **kwargs):
        return NotImplemented

    @abstractmethod
    def delete(self, where: str):
        return NotImplemented

    @abstractmethod
    def upsert(self, data: list[dict], **kwargs):
        return NotImplemented

    @abstractmethod
    def overwrite(self, data: list[dict] | DataFrame, **kwargs):
        return NotImplemented

    @abstractmethod
    def query(self, expr: str, **kwargs):
        return NotImplemented


class BaseNoSQLDataBase(BaseDataBase, ABC):
    _db_type: NoSQLDBType

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def create_table(self, table_name, **kwargs):
        return NotImplemented

    @abstractmethod
    def select(self, query: dict = None):
        return NotImplemented



from pydantic import BaseModel, Field, ConfigDict

from aa_rag import setting
from aa_rag.engine.lightrag import LightRAGInitParams, LightRAGIndexParams
from aa_rag.engine.simple_chunk import SimpleChunkIndexParams, SimpleChunkInitParams
from aa_rag.gtypes.enums import EngineType
from aa_rag.gtypes.models.base import BaseResponse
from aa_rag.gtypes.models.parse import ParserNeedItem


class BaseIndexItem(ParserNeedItem):
    pass


class IndexItem(BaseIndexItem):
    engine_type: EngineType = Field(
        default=setting.engine.type, examples=[setting.engine.type]
    )

    model_config = ConfigDict(extra="allow")


class SimpleChunkIndexItem(
    SimpleChunkInitParams, SimpleChunkIndexParams, BaseIndexItem
):
    source_data: None = Field(
        None, exclude=True, validate_default=False, deprecated=True
    )


class LightRAGIndexItem(LightRAGInitParams, LightRAGIndexParams, BaseIndexItem):
    source_data: None = Field(
        None, exclude=True, validate_default=False, deprecated=True
    )


class IndexResponse(BaseResponse):
    class Data(BaseModel):
        pass

    message: str = Field(
        default="Indexing completed via SimpleChunkIndex",
        examples=["Indexing completed via SimpleChunkIndex"],
    )
    data: Data = Field(default_factory=Data)



from os import PathLike
from pathlib import Path
from typing import Literal

from langchain_core.documents import Document
from markitdown import MarkItDown
from openai import OpenAI

from aa_rag import setting
from aa_rag.gtypes.enums import ParsingType
from aa_rag.oss import OSSResourceInfo
from aa_rag.parse.base import BaseParser


class MarkitDownParser(BaseParser):
    def __init__(
        self,
        use_cache: bool = True,
        update_cache: bool = True,
        llm: str = setting.llm.multimodal_model,
        **kwargs,
    ):
        """
        Initialize the MarkitDownParser with cache options and LLM settings.

        Args:
            use_cache (bool, optional): Whether to use cache. Defaults to True.
            update_cache (bool, optional): Whether to update cache. Defaults to True.
            llm (str, optional): The LLM model to use. Defaults to setting.llm.multimodal_model.
            **kwargs: Additional keyword arguments.
        """
        self.mtd_client = MarkItDown(
            llm_client=OpenAI(
                base_url=setting.openai.base_url, api_key=setting.openai.api_key
            ),
            llm_model=llm,
        )

        super().__init__(use_cache=use_cache, update_cache=update_cache)

    def type(self):
        """
        Return the type of the parser.

        Returns:
            ParsingType: The type of the parser.
        """
        return ParsingType.MARKITDOWN

    def _parse_file(
        self,
        file_path: PathLike,
        source: Literal["local", "oss"],
        oss_resource_info: OSSResourceInfo,
        **kwargs,
    ) -> Document:
        """
        Parse a file into a Document.

        Args:
            file_path (PathLike): The file path to parse.
            source (Literal["local", "oss"]): The source of the file.
            oss_resource_info (OSSResourceInfo): The OSS resource info.
            **kwargs: Additional keyword arguments.

        Returns:
            Document: The parsed document.
        """
        if Path(file_path).suffix in [".md"]:
            with open(file_path, mode="r") as file:
                content_str = file.read()
        else:
            content_str = self.mtd_client.convert(str(file_path)).text_content

        return Document(
            page_content=content_str,
            metadata={
                "source": f"{source}://{oss_resource_info.source_file_path}",
                "oss_info": oss_resource_info.model_dump(
                    exclude={
                        "url",
                        "suffix",
                    }
                ),
            },
        )

    def _parse_content(self, content: str, **kwargs) -> Document:
        """
        Parse content into a Document.

        Args:
            content (str): The content to parse.
            **kwargs: Additional keyword arguments.

        Returns:
            Document: The parsed document.
        """
        return Document(page_content=content)



from pathlib import Path

from tinydb import TinyDB, Query
from tinydb.table import Table

from aa_rag import setting
from aa_rag.db.base import BaseNoSQLDataBase, singleton
from aa_rag.gtypes.enums import NoSQLDBType


@singleton
class TinyDBDataBase(BaseNoSQLDataBase):
    table: Table | None = None
    _db_type = NoSQLDBType.TINYDB

    def __init__(self, uri: str = setting.storage.tinydb.uri, **kwargs):
        self.uri = uri
        # create parent directory if not exist
        Path(self.uri).parent.mkdir(parents=True, exist_ok=True)
        super().__init__(**kwargs)

    @property
    def connection(self) -> TinyDB:
        return self._conn_obj

    def connect(self):
        return TinyDB(self.uri)

    def create_table(self, table_name, **kwargs):
        return self.connection.table(table_name, **kwargs)

    def using(self, table_name, **kwargs):
        """
        get table object by table name, return self for with statement.
        """
        self.table = self.connection.table(table_name, **kwargs)
        return self

    def table_list(self):
        """
        return all table names in the database.
        """
        return self.connection.tables()

    def drop_table(self, table_name):
        """
        drop table by table name.
        """
        return self.connection.drop_table(table_name)

    def insert(self, data):
        """
        insert data into table.
        """
        return self.table.insert(data)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.table = None
        return False

    def close(self):
        self.connection.close()

    def _build_query_mongo(self, mongo_query: dict, q: Query = None):
        """
        Convert MongoDB query syntax to TinyDB query expressions.

        Supported MongoDB query operators include:
          - Comparison operators: $gt, $gte, $lt, $lte, $eq, $ne
          - Collection operators: $in, $nin
          - Existence judgment: $exists
          - Logical operators: $and, $or, $nor, $not

        Example:
          {"age": {"$gt": 30, "$lt": 50}, "name": "John"}
          Equivalent to:
          (Query()['age'] > 30) & (Query()['age'] < 50) & (Query()['name'] == "John")
        """
        if q is None:
            q = Query()
        if isinstance(mongo_query, dict):
            # handle top-level logical operators
            if "$and" in mongo_query:
                sub_exprs = [
                    self._build_query_mongo(sub, q) for sub in mongo_query["$and"]
                ]
                expr = sub_exprs[0]
                for e in sub_exprs[1:]:
                    expr = expr & e
                return expr
            elif "$or" in mongo_query:
                sub_exprs = [
                    self._build_query_mongo(sub, q) for sub in mongo_query["$or"]
                ]
                expr = sub_exprs[0]
                for e in sub_exprs[1:]:
                    expr = expr | e
                return expr
            elif "$nor" in mongo_query:
                sub_exprs = [
                    self._build_query_mongo(sub, q) for sub in mongo_query["$nor"]
                ]
                expr = sub_exprs[0]
                for e in sub_exprs[1:]:
                    expr = expr | e
                return ~expr
            elif "$not" in mongo_query:
                sub_expr = self._build_query_mongo(mongo_query["$not"], q)
                return ~sub_expr
            else:
                # handle field-level operators
                sub_exprs = []
                for field, condition in mongo_query.items():
                    if isinstance(condition, dict):
                        field_expr = None
                        for op, val in condition.items():
                            if op == "$gt":
                                current = q[field] > val
                            elif op == "$gte":
                                current = q[field] >= val
                            elif op == "$lt":
                                current = q[field] < val
                            elif op == "$lte":
                                current = q[field] <= val
                            elif op == "$eq":
                                current = q[field] == val
                            elif op == "$ne":
                                current = ~(q[field] == val)
                            elif op == "$in":
                                if not isinstance(val, list):
                                    raise ValueError("$in requires a list value")
                                # use the test method to construct in query
                                current = q[field].test(lambda v, lst=val: v in lst)
                            elif op == "$nin":
                                if not isinstance(val, list):
                                    raise ValueError("$nin requires a list value")
                                current = ~(q[field].test(lambda v, lst=val: v in lst))
                            elif op == "$exists":
                                # check the field is None if val is False, otherwise check it is not None
                                current = (
                                    (q[field] is not None)
                                    if val
                                    else (q[field] is None)
                                )
                            else:
                                raise ValueError(f"Unsupported Operation: {op}")
                            if field_expr is None:
                                field_expr = current
                            else:
                                field_expr = field_expr & current
                        sub_exprs.append(field_expr)
                    else:
                        # make a direct judgment of equal value
                        sub_exprs.append(q[field] == condition)
                if not sub_exprs:
                    return None
                expr = sub_exprs[0]
                for e in sub_exprs[1:]:
                    expr = expr & e
                return expr
        else:
            raise ValueError("Mongo Query Condition Must Be A Dict")

    def select(self, query: dict = None):
        """
        Query the interface and receive MongoDB query syntax.
        Args:
            query: A dict representing the MongoDB query.

        Example:
            query = {"age": {"$gt": 30}, "$or": [{"name": "Alice"}, {"name": "Bob"}]}
            results = db.select(query)
        """
        if query is None or not query:
            return self.table.all()
        query_obj = self._build_query_mongo(query)
        return self.table.search(query_obj)

    def update(self, update_data: dict, query: dict = None):
        """
        Update records that meet MongoDB query criteria.

        Args:
            update_data: A dictionary that specifies the fields and values to be updated.
            query: MongoDB dictionary of query syntax, used to select records that need to be updated.
                   If query is None or an empty dictionary, all records are updated.

        Returns:
            update_result: Usually an updated record identification list.
        """
        if query is None or not query:
            # If no query conditions are specified, all records will be updated.
            return self.table.update(update_data)
        query_obj = self._build_query_mongo(query)
        return self.table.update(update_data, query_obj)

    def delete(self, query: dict = None):
        """
        Delete records that meet MongoDB query criteria.

        Args:
            query: MongoDB dictionary of query syntax, used to select records to be deleted.
                   If query is None or an empty dictionary, all records are deleted.

        Returns:
            delete_result: Usually a list of deleted records.
        """
        if query is None or not query:
            # If no query conditions are specified, all records will be deleted.
            return self.table.remove()
        query_obj = self._build_query_mongo(query)
        return self.table.remove(query_obj)



from typing import List

from langchain_core.documents import Document
from pydantic import BaseModel, Field, ConfigDict, field_validator

from aa_rag import setting
from aa_rag.engine.lightrag import LightRAGInitParams, LightRAGRetrieveParams
from aa_rag.engine.simple_chunk import SimpleChunkRetrieveParams, SimpleChunkInitParams
from aa_rag.gtypes.enums import EngineType
from aa_rag.gtypes.models.base import BaseResponse


class BaseRetrieveItem(BaseModel):
    pass


class RetrieveItem(BaseModel):
    engine_type: EngineType = Field(
        default=setting.engine.type, examples=[setting.engine.type]
    )

    model_config = ConfigDict(extra="allow")


class SimpleChunkRetrieveItem(
    SimpleChunkInitParams, SimpleChunkRetrieveParams, BaseRetrieveItem
):
    pass


class LightRAGRetrieveItem(
    LightRAGInitParams, LightRAGRetrieveParams, BaseRetrieveItem
):
    pass


class RetrieveResponse(BaseResponse):
    class Data(BaseModel):
        documents: List[Document] = Field(default=..., examples=[[]])

        @field_validator("documents")
        def validate_documents(cls, v):
            doc_s = [doc.model_dump(include={"metadata", "page_content"}) for doc in v]

            for doc in v:
                if "metadata" in doc:
                    if "identifier" in doc["metadata"]:
                        doc.pop("identifier")

            return doc_s

    message: str = Field(
        default="Retrieval completed via BaseRetrieve", examples=["Retrieval completed"]
    )
    data: Data = Field(default_factory=Data)



import pandas as pd
import json
import argparse


# 示例原始数据格式可能类似：
# df = pd.DataFrame({
#     'url': ['x1', 'x1', 'x1', 'x2'],
#     'name': ['n1', 'n1', 'n1', 'n2'],
#     'os': ['win', 'win', 'linux', 'mac'],
#     'arch': ['amd64', 'amd64', 'arm', 'arm64'],
#     'procedure': ['p1', 'p1', 'p2', 'p3'],
#     'question': ['q1', 'q2', 'q3', 'q4'],
#     'answer': ['a1', 'a2', 'a3', 'a4']
# })


def main():
    # 设置命令行参数
    parser = argparse.ArgumentParser(description="Convert CSV to deployment JSON")
    parser.add_argument("--input", required=True, help="Input CSV file path")
    args = parser.parse_args()

    # 读取CSV文件
    df = pd.read_csv(args.input)
    df.fillna("", inplace=True)

    # 按url和name分组聚合
    grouped = df.groupby(["url", "name"])

    result = []
    for (url, name), group in grouped:
        # 提取部署配置（去重）
        deployments = group[["os", "arch", "procedure"]].drop_duplicates()

        # 提取QA对（去重）
        qas = group[["question", "answer"]].drop_duplicates()

        # 构建条目
        entry = {
            "url": url,
            "name": name,
            "deployment": deployments.to_dict("records"),
            "qa": qas.to_dict("records"),
        }
        result.append(entry)

    # 生成带缩进的JSON
    json_output = json.dumps(result, indent=2, ensure_ascii=False)

    # 保存到文件
    with open("deployments.json", "w") as f:
        f.write(json_output)


if __name__ == "__main__":
    main()



from configparser import ConfigParser
from typing import List, Union, Literal, Dict

from langchain_core.documents import Document
from lightrag import LightRAG, QueryParam
from lightrag.llm.openai import openai_embed, openai_complete
from pydantic import BaseModel, Field, SecretStr

from aa_rag import setting, utils
from aa_rag.db.base import BaseNoSQLDataBase
from aa_rag.engine.base import BaseEngine
from aa_rag.gtypes.enums import EngineType

dfs_setting = setting.engine.lightrag


# 参数模型定义
class LightRAGInitParams(BaseModel):
    knowledge_name: str = Field(..., description="The name of the knowledge")
    identifier: str = Field(
        default="common", description="The identifier of the knowledge"
    )

    llm: str = Field(default=dfs_setting.llm, description="The language model to use.")


class LightRAGIndexParams(BaseModel):
    source_data: Union[Document, List[Document]] = Field(
        ..., description="The source data to index."
    )


class LightRAGRetrieveParams(BaseModel):
    query: str = Field(..., description="The query string to retrieve.")
    retrieve_mode: Literal["local", "global", "hybrid", "naive", "mix"] = "hybrid"
    """Specifies the retrieval mode:
    - "local": Focuses on context-dependent information.
    - "global": Utilizes global knowledge.
    - "hybrid": Combines local and global retrieval methods.
    - "naive": Performs a basic search without advanced techniques.
    - "mix": Integrates knowledge graph and vector retrieval.
    """
    top_k: int = Field(
        default=dfs_setting.k,
        description="Number of top items to retrieve. Represents entities in 'local' mode and relationships in 'global' mode.",
    )


class LightRAGGenerateParams(BaseModel):
    pass


class LightRAGEngine(
    BaseEngine[LightRAGIndexParams, LightRAGRetrieveParams, LightRAGGenerateParams]
):
    def __init__(
        self,
        params: LightRAGInitParams,
        embedding_model: str = dfs_setting.embedding.model,
        **kwargs,
    ):
        namespace_prefix = (
            f"{params.knowledge_name}_{embedding_model}_{params.identifier}".replace(
                "-", "_"
            )
        )

        self._generate_ini_config_file()  # !! generate config.ini file because lightrag do not support specifying config file path

        self.rag = LightRAG(
            working_dir=dfs_setting.dir,
            embedding_func=openai_embed,
            llm_model_name=params.llm,
            llm_model_func=openai_complete,
            namespace_prefix=namespace_prefix,
            addon_params={}
            if kwargs.get("addon_params") is None
            else kwargs.get("addon_params"),
            vector_storage=dfs_setting.vector_storage.value,
            graph_storage=dfs_setting.graph_storage.value,
            vector_db_storage_cls_kwargs={
                "cosine_better_than_threshold": dfs_setting.cosine_threshold
            },
            log_level=10,  # DEBUG Level
        )

        # get nosql obj to store the source data
        self.db: BaseNoSQLDataBase = utils.get_nosql_db(setting.storage.nosql)
        self.table_name = namespace_prefix

    @property
    def type(self):
        """
        Return the type of the engine.
        """
        return EngineType.LightRAG

    @staticmethod
    def _generate_ini_config_file():
        config = ConfigParser()

        milvus_config_dict = setting.storage.milvus.model_dump()
        config.add_section("milvus")
        for k, v in milvus_config_dict.items():
            if v:
                if isinstance(v, SecretStr):
                    v = v.get_secret_value()
                if k == "db_name":
                    v = v + "_lightrag"
                config.set("milvus", k, v)

        neo4j_config_dict = setting.storage.neo4j.model_dump()
        config.add_section("neo4j")
        for k, v in neo4j_config_dict.items():
            if v:
                if isinstance(v, SecretStr):
                    v = v.get_secret_value()
                config.set("neo4j", k, v)

        with open("config.ini", "w") as f:
            config.write(f)

    async def index(self, params: LightRAGIndexParams):
        if self.table_name not in self.db.table_list():
            self.db.create_table(self.table_name)

        id_s = []

        with self.db.using(self.table_name) as table:
            docs = (
                [params.source_data]
                if isinstance(params.source_data, Document)
                else params.source_data
            )
            for doc in docs:
                doc_id = utils.calculate_md5(doc.page_content)
                id_s.append(doc_id)
                table.insert({"doc_id": doc_id, **doc.model_dump()})

        await self.rag.ainsert(input=[doc.page_content for doc in docs], ids=id_s)

    async def retrieve(self, params: LightRAGRetrieveParams):
        context_str = await self.rag.aquery(
            query=params.query,
            param=QueryParam(
                mode=params.retrieve_mode, only_need_context=True, top_k=params.top_k
            ),
        )
        entity_df, rel_df = utils.markdown_extract_csv_df(context_str)

        metadata_s = await self._get_doc_metadata(entity_df, rel_df)

        return [
            Document(
                page_content=context_str,
                metadata={"source": metadata_s},
            )
        ]

    def generate(self, params: LightRAGGenerateParams):
        pass

    async def _get_doc_metadata(self, entity_df, rel_df) -> List[Dict]:
        chunk_id_s = set()
        metadata_s: List = list()  # file source
        for entity in entity_df["entity"].values:
            entity: str = entity.strip().replace("'", "").replace('"', "")
            entity_info = await self.rag.get_entity_info(entity)
            if entity_info:
                source_info = await self.rag.text_chunks.get_by_id(
                    entity_info["source_id"]
                )
                if source_info:
                    chunk_id_s.add(source_info["full_doc_id"])

        for _, rel in rel_df.iterrows():
            source: str = rel["source"].strip().replace("'", "").replace('"', "")
            target: str = rel["target"].strip().replace("'", "").replace('"', "")
            rel_info = await self.rag.get_relation_info(source, target)
            if rel_info["source_id"]:
                source_info = await self.rag.text_chunks.get_by_id(
                    rel_info["source_id"]
                )
                if source_info:
                    chunk_id_s.add(source_info["full_doc_id"])

        with self.db.using(self.table_name) as table:
            docs: List[Dict] = table.select(query={"doc_id": {"$in": list(chunk_id_s)}})
            metadata_s.append(doc.get("metadata") for doc in docs)
            if None in metadata_s:
                metadata_s.remove(None)

        return metadata_s



from fastapi import APIRouter, HTTPException

from aa_rag import utils
from aa_rag.engine.lightrag import (
    LightRAGEngine,
    LightRAGInitParams,
    LightRAGIndexParams,
)
from aa_rag.engine.simple_chunk import (
    SimpleChunk,
    SimpleChunkInitParams,
    SimpleChunkIndexParams,
)
from aa_rag.gtypes.enums import EngineType
from aa_rag.gtypes.models.index import (
    IndexItem,
    SimpleChunkIndexItem,
    IndexResponse,
    LightRAGIndexItem,
)
from aa_rag.gtypes.models.parse import ParserNeedItem

router = APIRouter(
    prefix="/index", tags=["Index"], responses={404: {"description": "Not found"}}
)


@router.post("/")
async def root(item: IndexItem):
    match item.engine_type:
        case EngineType.SimpleChunk:
            chunk_item = SimpleChunkIndexItem(**item.model_dump())
            return await chunk_index(chunk_item)
        case _:
            raise HTTPException(status_code=400, detail="IndexType not supported")


@router.post("/chunk", tags=["SimpleChunk"])
async def chunk_index(item: SimpleChunkIndexItem) -> IndexResponse:
    source_data = await utils.parse_content(params=ParserNeedItem(**item.model_dump()))

    # index content
    engine = SimpleChunk(params=SimpleChunkInitParams(**item.model_dump()))

    engine.index(
        params=SimpleChunkIndexParams(
            **{
                **item.model_dump(),
                "source_data": source_data,
            }
        )
    )

    return IndexResponse(
        code=200,
        status="success",
        message="Indexing completed via SimpleChunkIndex",
        data=IndexResponse.Data(),
    )


@router.post("/lightrag", tags=["LightRAG"])
async def lightrag_index(item: LightRAGIndexItem) -> IndexResponse:
    # parse content
    source_data = await utils.parse_content(params=ParserNeedItem(**item.model_dump()))

    # index content
    engine = LightRAGEngine(params=LightRAGInitParams(**item.model_dump()))

    await engine.index(
        params=LightRAGIndexParams(
            **{
                **item.model_dump(),
                "source_data": source_data,
            }
        )
    )

    return IndexResponse(
        code=200,
        status="success",
        message="Indexing completed via LightRAGIndex",
        data=IndexResponse.Data(),
    )



from fastapi import APIRouter

from aa_rag.gtypes.models.knowlege_base.qa import (
    QAIndexItem,
    QAIndexResponse,
    QARetrieveItem,
    QARetrieveResponse,
)
from aa_rag.knowledge_base.built_in.qa import QAKnowledge

router = APIRouter(
    prefix="/qa", tags=["qa"], responses={404: {"description": "Not Found"}}
)


@router.get("/")
async def root():
    return {
        "built_in": True,
        "description": "问题/解决方案库",
    }


@router.post("/index")
async def index(item: QAIndexItem):
    qa = QAKnowledge()

    result = qa.index(
        **item.model_dump(include={"error_desc", "error_solution", "tags"})
    )

    return QAIndexResponse(
        code=200,
        status="success",
        message="Indexing completed in QA Knowledge Base",
        data=QAIndexResponse.Data(affect_row_ids=result),
    )


@router.post("/retrieve")
async def retrieve(item: QARetrieveItem):
    qa = QAKnowledge()

    result = qa.retrieve(**item.model_dump(include={"error_desc", "tags"}))
    return QARetrieveResponse(
        code=200,
        status="success",
        message="Retrieving completed in QA Knowledge Base",
        data=QARetrieveResponse.Data(qa=result),
    )



from abc import abstractmethod
from typing import Any, List

from aa_rag import setting, utils


class BaseKnowledge:
    dimensions: int

    def __init__(
        self,
        llm: str = setting.llm.model,
        embedding_model: str = setting.embedding.model,
        **kwargs,
    ):
        self.llm = utils.get_llm(llm)
        self.embedding_model, self.dimensions = utils.get_embedding_model(
            embedding_model, return_dim=True
        )

    @property
    @abstractmethod
    def knowledge_name(self):
        return NotImplemented

    def index(self, **kwargs):
        return NotImplemented

    def retrieve(self, **kwargs) -> List[Any]:
        return NotImplemented



from typing import Optional

from pydantic import Field

from aa_rag.oss import OSSStoreInitParams


class StoreImageParams(OSSStoreInitParams):
    image: Optional[str] = Field(default=None, description="Base64 encoded image.")
    img_desc: Optional[str] = Field(
        default=None, description="The description of image."
    )



from fastapi import APIRouter

from aa_rag.gtypes.models.knowlege_base.solution import (
    SolutionIndexItem,
    SolutionIndexResponse,
    SolutionRetrieveItem,
    Guide,
    SolutionRetrieveResponse,
)
from aa_rag.knowledge_base.built_in.solution import SolutionKnowledge

router = APIRouter(
    prefix="/solution", tags=["solution"], responses={404: {"description": "Not Found"}}
)


@router.get("/")
async def root():
    return {
        "built_in": True,
        "description": "项目部署方案库",
    }


@router.post("/index")
async def index(item: SolutionIndexItem):
    solution = SolutionKnowledge(**item.model_dump(include={"llm", "embedding_model"}))

    affect_row_num_s = solution.index(
        **item.model_dump(include={"env_info", "procedure", "project_meta"})
    )

    return SolutionIndexResponse(
        code=200,
        status="success",
        data=SolutionIndexResponse.Data(
            affect_row_num=affect_row_num_s, table_name="solution"
        ),
    )


@router.post("/retrieve")
async def retrieve(item: SolutionRetrieveItem):
    solution = SolutionKnowledge(
        **item.model_dump(include={"llm", "embedding_model", "relation_db_path"})
    )

    guide: Guide | None = solution.retrieve(
        **item.model_dump(include={"env_info", "project_meta"})
    )

    return SolutionRetrieveResponse(
        code=200, status="success", data=SolutionRetrieveResponse.Data(guide=guide)
    )



from enum import Enum


class IndexType(Enum):
    CHUNK: str = "chunk"

    def __str__(self):
        return f"{self.value}"


class RetrieveType(Enum):
    HYBRID: str = "hybrid"
    DENSE: str = "dense"
    BM25: str = "bm25"

    def __str__(self):
        return f"{self.value}"


class DBMode(Enum):
    INSERT = "insert"
    OVERWRITE = "overwrite"
    UPSERT = "upsert"

    def __str__(self):
        return f"{self.value}"


class VectorDBType(Enum):
    LANCE: str = "lance"
    MILVUS: str = "milvus"

    def __str__(self):
        return f"{self.value}"


class NoSQLDBType(Enum):
    TINYDB: str = "tinydb"
    MONGODB: str = "mongodb"

    def __str__(self):
        return f"{self.value}"


class EngineType(Enum):
    SimpleChunk: str = "chunk"
    LightRAG: str = "lightrag"

    def __str__(self):
        return f"{self.value}"


class ParsingType(Enum):
    MARKITDOWN: str = "markitdown"

    def __str__(self):
        return f"{self.value}"


class LightRAGVectorStorageType(Enum):
    MILVUS: str = "MilvusVectorDBStorage"

    def __str__(self):
        return f"{self.value}"


class LightRAGGraphStorageType(Enum):
    NEO4J: str = "Neo4JStorage"
    NETWORKX: str = "NetworkXStorage"

    def __str__(self):
        return f"{self.value}"



import copy
from pathlib import Path
from typing import List, Union, Dict

from pandas import DataFrame
from pymilvus import (
    CollectionSchema,
    MilvusClient,
    DataType,
)

from aa_rag import setting
from aa_rag.db.base import BaseVectorDataBase, singleton
from aa_rag.gtypes.enums import VectorDBType


@singleton
class MilvusDataBase(BaseVectorDataBase):
    """Milvus vector database implementation with unified interface"""

    _db_type = VectorDBType.MILVUS
    using_collection_name: str | None = None

    def __init__(
        self,
        uri: str = setting.storage.milvus.uri,
        user: str = setting.storage.milvus.user,
        password: str = setting.storage.milvus.password.get_secret_value(),
        db_name: str = setting.storage.milvus.db_name,
        **kwargs,
    ):
        # create parent directory if not exist
        if uri.startswith("http"):
            uri = uri
        else:
            Path(uri).parent.mkdir(parents=True, exist_ok=True)
        super().__init__(
            uri=uri, user=user, password=password, db_name=db_name, **kwargs
        )

    @property
    def connection(self) -> MilvusClient:
        return self._conn_obj

    def connect(self, **kwargs) -> MilvusClient:
        """Connect to Milvus server"""
        return MilvusClient(**kwargs)

    def table_list(self, **kwargs) -> List[str]:
        """List all collection names"""
        return self.connection.list_collections()

    def create_table(self, table_name: str, schema: CollectionSchema, **kwargs):
        """Create new collection with schema"""
        if table_name not in self.table_list():
            if kwargs.get("index_params"):
                self.connection.create_collection(
                    collection_name=table_name,
                    schema=schema,
                    index_params=kwargs.get("index_params"),
                )
            else:
                index_params = self.connection.prepare_index_params()
                index_params.add_index(
                    field_name="vector", index_type="AUTOINDEX", metric_type="L2"
                )
                self.connection.create_collection(
                    collection_name=table_name, schema=schema, index_params=index_params
                )
        else:
            raise ValueError(f"Collection {table_name} already exists")

    def drop_table(self, table_name: str):
        """Drop specified collection"""
        self.connection.drop_collection(table_name)

    def using(self, collection_name: str, **kwargs):
        """Set table to use"""
        self.connection.load_collection(collection_name=collection_name)
        self.using_collection_name = collection_name

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # self.connection.release_collection(collection_name=self.using_collection_name)
        self.using_collection_name = None
        return False

    def close(self):
        self.connection.close()

    def insert(self, data: List[dict], **kwargs):
        """Insert data into collection"""
        assert self.using_collection_name, "Collection not loaded. Use using() first"
        res = self.connection.insert(
            collection_name=self.using_collection_name, data=data
        )
        return res

    def delete(self, where: str, **kwargs):
        """Delete entities with boolean expression"""
        assert self.using_collection_name, "Collection not loaded. Use using() first"
        return self.connection.delete(self.using_collection_name, where, **kwargs)

    def upsert(self, data: list[dict] | DataFrame, **kwargs):
        """Upsert data into collection with JSON merge support"""
        assert self.using_collection_name, "Collection not loaded. Use using() first"

        def update_old_json_with_new(new_data):
            # 获取字段描述
            fields = self.connection.describe_collection(self.using_collection_name)[
                "fields"
            ]

            # 健壮的主键获取（修复问题4）
            primary_keys = [
                field["name"] for field in fields if field.get("is_primary", False)
            ]
            if not primary_keys:
                raise ValueError("Collection has no primary key")
            primary_key = primary_keys[0]

            json_fields = [
                field["name"] for field in fields if field["type"] == DataType.JSON
            ]

            # 获取所有需要更新的字段（修复问题3）
            all_fields = list(
                {primary_key}.union(json_fields, *[d.keys() for d in new_data])
            )

            # 构建查询表达式（修复潜在的类型问题）
            pk_values = [str(d[primary_key]) for d in new_data if primary_key in d]
            if not pk_values:
                return new_data  # 无主键直接插入新数据

            # 查询旧数据（包含所有必要字段）
            old_data = self.query(
                expr=f"{primary_key} in {pk_values}", output_fields=all_fields
            )
            old_data_map = {str(d[primary_key]): d for d in old_data}

            # 递归合并函数（修复问题2）
            def deep_merge(
                original: Union[Dict, List, None], new: Union[Dict, List]
            ) -> Union[Dict, List]:
                if original is None:
                    original = {} if isinstance(new, dict) else []
                merged = copy.deepcopy(original)
                if isinstance(merged, dict) and isinstance(new, dict):
                    for k, v in new.items():
                        if k not in merged:
                            merged[k] = copy.deepcopy(v)
                        elif isinstance(v, dict) and isinstance(merged[k], dict):
                            merged[k] = deep_merge(merged[k], v)
                        elif isinstance(v, list) and isinstance(merged[k], list):
                            merged[k] = list(set(merged[k] + v))  # 保留顺序
                        else:
                            merged[k] = v
                elif isinstance(merged, list) and isinstance(new, list):
                    merged = list(set(merged + new))
                else:
                    raise ValueError(f"Cannot merge {type(original)} with {type(new)}")
                return merged

            # 合并数据（修复问题1）
            merged_data = []
            for new_item in new_data:
                pk_value = str(new_item.get(primary_key))
                old_item = old_data_map.get(pk_value, {})

                # only merge json fields
                for field in json_fields:
                    new_item[field] = deep_merge(
                        old_item.get(field), new_item.get(field)
                    )

                merged_data.append(new_item)

            return merged_data

        # 数据格式转换
        if isinstance(data, DataFrame):
            data = data.to_dict(orient="records")

        # 执行合并逻辑
        processed_data = update_old_json_with_new(data)

        # 执行upsert
        return self.connection.upsert(
            collection_name=self.using_collection_name, data=processed_data
        )

    def overwrite(self, data: list[dict] | DataFrame, **kwargs):
        assert self.using_collection_name, "Collection not loaded. Use using() first"
        self.delete(where="id is not null")  # truncate collection
        return self.insert(data=data)

    # def search(
    #         self,
    #         query_vector: List[float],
    #         top_k: int = 3,
    #         anns_field: str = "vector",
    #         **kwargs,
    # ):
    #     """Vector similarity search"""
    #
    #     # Convert to numpy array and normalize if needed
    #     vector = np.array(query_vector, dtype=np.float32)
    #
    #     res = self.connection.search(
    #         collection_name=self.using_collection_name,
    #         anns_field=anns_field,
    #         data=[vector],
    #         limit=top_k,
    #         search_params={"metric_type": "IP"},
    #         **kwargs,
    #     )
    #
    #     return [
    #         {"id": hit.id, "distance": hit.distance, **hit.entity.to_dict()}
    #         for hit in res[0]
    #     ]

    def query(self, expr, **kwargs):
        iterator = self.connection.query_iterator(
            batch_size=1000,
            collection_name=self.using_collection_name,
            filter=expr,
            output_fields=kwargs.get("output_fields", None),
            limit=kwargs.get("limit", -1),
        )

        results = []

        while True:
            result = iterator.next()
            if not result:
                iterator.close()
                break

            results += result

        return results


if __name__ == "__main__":
    milvus_db = MilvusDataBase()
    milvus_db.connect()
    print(milvus_db.table_list())

    with milvus_db.using("user_guide_chunk_text_embedding_3_small") as db:
        milvus_db.query(
            "id in ['8672a4c387ff30688588e22f2e5e7c6c']",
            limit=10,
            output_fields=["id", "text"],
        )



from typing import List, Any

from langchain_core.documents import Document

from aa_rag import setting
from aa_rag.engine.simple_chunk import (
    SimpleChunk,
    SimpleChunkInitParams,
    SimpleChunkIndexParams,
    SimpleChunkRetrieveParams,
)
from aa_rag.gtypes.enums import VectorDBType
from aa_rag.knowledge_base.base import BaseKnowledge


class QAKnowledge(BaseKnowledge):
    def __init__(
        self,
        vector_db: VectorDBType = setting.storage.vector,
        **kwargs,
    ):
        """
        QA Knowledge Base. Built-in Knowledge Base.
        """
        super().__init__(**kwargs)

        # define table schema
        if vector_db == VectorDBType.LANCE:
            import pyarrow as pa

            schema = pa.schema(
                [
                    pa.field("id", pa.utf8(), False),
                    pa.field("vector", pa.list_(pa.float64(), self.dimensions), False),
                    pa.field("text", pa.utf8(), False),
                    pa.field(
                        "metadata",
                        pa.struct(
                            [
                                pa.field("solution", pa.utf8(), False),
                                pa.field("tags", pa.list_(pa.utf8()), False),
                            ]
                        ),
                        False,
                    ),
                ]
            )
        else:
            schema = None

        self.engine = SimpleChunk(
            params=SimpleChunkInitParams(knowledge_name=self.knowledge_name.lower()),
            db_type=vector_db,
            schema=schema,
        )
        self.db = vector_db

    @property
    def knowledge_name(self):
        return "QA"

    def index(self, error_desc: str, error_solution: str, tags: List[str], **kwargs):
        """
        Index the QA information.
        Args:
            error_desc: The error description.
            error_solution: The solution of the QA.
            tags: The tags of the QA.
            **kwargs:
        # check if the project is already indexed
        """

        chunk_size = (
            len(error_desc) * 2
            if kwargs.get("chunk_size") is None
            else kwargs.get("chunk_size")
        )  # ensure the chunk size is large enough to cover the whole text. do not split the text.
        chunk_overlap = (
            0 if kwargs.get("chunk_overlap") is None else kwargs.get("chunk_overlap")
        )

        self.engine.index(
            params=SimpleChunkIndexParams(
                source_data=Document(
                    page_content=error_desc,
                    metadata={"solution": error_solution, "tags": tags},
                ),
                chunk_size=chunk_size,
                chunk_overlap=chunk_overlap,
            )
        )

    def retrieve(self, error_desc: str, tags: List[str] = None) -> List[Any]:
        """
        Retrieve the QA information.
        Args:
            error_desc: The error description.
            tags: The tags of the QA.
        Returns:
            List[Any]: The QA information.
        """
        return self.engine.retrieve(
            SimpleChunkRetrieveParams(query=error_desc, top_k=1)
        )



from abc import abstractmethod
from typing import TypeVar, Generic

from pydantic import BaseModel

# 定义泛型参数
IndexT = TypeVar("IndexT", bound=BaseModel)
RetrieveT = TypeVar("RetrieveT", bound=BaseModel)
GenerateT = TypeVar("GenerateT", bound=BaseModel)


class BaseEngine(Generic[IndexT, RetrieveT, GenerateT]):
    @property
    @abstractmethod
    def type(self):
        """
        Return the type of the engine.
        """
        ...

    @abstractmethod
    def index(self, params: IndexT):
        """
        Build index from source data and store to database.
        """
        ...

    @abstractmethod
    def retrieve(self, params: RetrieveT):
        """
        Retrieve data.
        """
        ...

    @abstractmethod
    def generate(self, params: GenerateT):
        """
        Generate data.
        """
        ...



import json
from typing import Dict, Any, List

from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate

from aa_rag import utils, setting
from aa_rag.db.base import BaseNoSQLDataBase
from aa_rag.gtypes.enums import NoSQLDBType
from aa_rag.gtypes.models.knowlege_base.solution import CompatibleEnv, Project, Guide
from aa_rag.knowledge_base.base import BaseKnowledge


class SolutionKnowledge(BaseKnowledge):
    @property
    def knowledge_name(self):
        return "Solution"

    def __init__(self, nosql_db: NoSQLDBType = setting.storage.nosql, **kwargs):
        """
        Initialize the SolutionKnowledge class.

        Args:
            nosql_db (NoSQLDBType): The type of NoSQL database to use (default is TinyDB).
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self.nosql_db: BaseNoSQLDataBase = utils.get_db(nosql_db)
        self.table_name = self.knowledge_name.lower()

    def _is_compatible_env(
        self, source_env_info: CompatibleEnv, target_env_info: CompatibleEnv
    ) -> bool:
        """
        Determine if source_env_info is compatible with target_env_info.

        Args:
            source_env_info (CompatibleEnv): The source environment information.
            target_env_info (CompatibleEnv): The target environment information.

        Returns:
            bool: True if compatible, False otherwise.
        """
        prompt_template = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    """You are an expert in computer hardware device information. 
                    I will provide you with two jsons. Each json is the detailed data of a computer hardware device information.
                    --Requirements--
                    1. Please determine whether the two devices are compatible. If compatible, please return "True". Otherwise, return "False".
                    2. Do not return other information. Just return "True" or "False".

                    --Data--
                    source_env_info: {source_env_info}
                    target_env_info: {target_env_info}

                    --Result--
                    result:
                    """,
                )
            ]
        )

        chain = prompt_template | self.llm | StrOutputParser()
        result = chain.invoke(
            {
                "source_env_info": json.dumps(source_env_info.model_dump()),
                "target_env_info": json.dumps(target_env_info.model_dump()),
            }
        )
        try:
            result = bool(result)
        except Exception:
            result = False
        return result

    def _get_project_in_db(self, project_meta: Dict[str, Any]) -> Project | None:
        """
        Retrieve a project record from TinyDB by project name and return a Project object.

        Args:
            project_meta (Dict[str, Any]): The project metadata.

        Returns:
            Project | None: The Project object if found, None otherwise.
        """
        query = {"name": project_meta["name"]}
        with self.nosql_db.using(self.table_name) as table:
            records = table.select(query)
        if records:
            record = records[0]
            guides_data: List[Dict[str, Any]] = record.get("guides", [])
            guides: List[Guide] = [
                Guide(
                    procedure=item["procedure"],
                    compatible_env=CompatibleEnv(**item["compatible_env"]),
                )
                for item in guides_data
            ]
            project_id = record.get("project_id", None)
            return Project(
                **record.get("project_meta", {}), guides=guides, id=project_id
            )
        else:
            return None

    def _project_to_db(self, project: Project) -> int:
        """
        Save a project to TinyDB. Insert if project.id is None, otherwise update.

        Args:
            project (Project): The project to save.

        Returns:
            int: 1 indicating success.
        """
        record = {
            "guides": [guide.model_dump() for guide in project.guides],
            "project_meta": project.model_dump(exclude={"guides", "id"}),
            "name": project.model_dump(exclude={"guides", "id"}).get("name"),
        }
        with self.nosql_db.using(self.table_name) as table:
            if project.id is None:
                project_id = utils.get_uuid()
                record["project_id"] = project_id
                table.insert(record)
            else:
                table.update(record, query={"project_id": project.id})
        return 1

    def _merge_procedure(self, source_procedure: str, target_procedure: str) -> str:
        """
        Merge source_procedure with target_procedure and return the merged procedure in MarkDown format.

        Args:
            source_procedure (str): The source procedure.
            target_procedure (str): The target procedure.

        Returns:
            str: The merged procedure in MarkDown format.
        """
        prompt_template = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    """Merge the source procedure with the target procedure.
                    --Requirements--
                    1. The merged procedure should be in a MarkDown format.
                    2. Just return the merged procedure. Do not return other information.
                    --Data--
                    source_procedure: {source_procedure}
                    target_procedure: {target_procedure}
                    --Result--
                    merged_procedure:
                    """,
                )
            ]
        )

        chain = prompt_template | self.llm | StrOutputParser()
        result: str = chain.invoke(
            {"source_procedure": source_procedure, "target_procedure": target_procedure}
        )

        return result

    def index(
        self, env_info: Dict[str, Any], procedure: str, project_meta: Dict[str, Any]
    ) -> int:
        """
        Write a solution to the knowledge base. Merge procedures if compatible, otherwise add a new guide.

        Args:
            env_info (Dict[str, Any]): The environment information.
            procedure (str): The procedure to index.
            project_meta (Dict[str, Any]): The project metadata.

        Returns:
            int: 1 indicating success.
        """
        env_info_obj = CompatibleEnv(**env_info)

        project = self._get_project_in_db(project_meta)
        if project:
            for guide in project.guides:
                is_compatible: bool = self._is_compatible_env(
                    env_info_obj, guide.compatible_env
                )
                if is_compatible:
                    merged_procedure = self._merge_procedure(guide.procedure, procedure)
                    guide.procedure = merged_procedure
                    break
            else:
                guide = Guide(procedure=procedure, compatible_env=env_info_obj)
                project.guides.append(guide)
        else:
            guide = Guide(procedure=procedure, compatible_env=env_info_obj)
            project = Project(guides=[guide], **project_meta)

        return self._project_to_db(project)

    def retrieve(
        self, env_info: Dict[str, Any], project_meta: Dict[str, Any]
    ) -> Guide | None:
        """
        Retrieve a guide from the knowledge base that is compatible with the given environment.

        Args:
            env_info (Dict[str, Any]): The environment information.
            project_meta (Dict[str, Any]): The project metadata.

        Returns:
            Guide | None: The compatible guide if found, None otherwise.
        """
        env_info_obj = CompatibleEnv(**env_info)
        project = self._get_project_in_db(project_meta)
        if project:
            for guide in project.guides:
                is_compatible: bool = self._is_compatible_env(
                    env_info_obj, guide.compatible_env
                )
                if is_compatible:
                    return guide
        return None



from typing import Any

from aa_rag import setting
from aa_rag.db.base import BaseNoSQLDataBase, singleton


@singleton
class MongoDBDataBase(BaseNoSQLDataBase):
    _db_type = "MongoDB"

    collection: Any

    def __init__(
        self,
        uri: str = setting.storage.mongodb.uri,
        db_name: str = setting.storage.mongodb.db_name,
        **kwargs,
    ):
        """
        Initialize the MongoDB database instance.

        Args:
            uri: MongoDB connection URI, e.g., "mongodb://localhost:27017"
            db_name: The database name to use; defaults to "default"
            **kwargs: Additional keyword arguments
        """

        super().__init__(uri=uri, db_name=db_name, **kwargs)

    def connect(self, **kwargs):
        """Establish and return a MongoDB client connection."""
        try:
            from pymongo import MongoClient
        except ImportError:
            raise ImportError(
                "MongoDB can only be enabled on the online service, please execute `pip install aa-rag[online]`."
            )
        client = MongoClient(kwargs.get("uri"))

        return client[kwargs.get("db_name")]

    def create_table(self, collection_name: str, **kwargs):
        """
        Create and return a collection with the specified name.
        If the collection already exists, return the existing one.

        Args:
            collection_name: The name of the collection.
            **kwargs: Additional optional parameters for collection creation.

        Returns:
            The collection object.
        """
        if collection_name in self.table_list():
            return self.connection[collection_name]
        else:
            # The create_collection method will raise an exception if the collection exists,
            # so we check for its existence first.
            return self.connection.create_collection(collection_name, **kwargs)

    def using(self, collection_name: str, **kwargs):
        """
        Set the current active collection and return self to support method chaining.

        Args:
            collection_name: The name of the collection.
            **kwargs: Reserved for additional parameters.

        Returns:
            self
        """
        self.collection = self.connection[collection_name]
        return self

    def table_list(self):
        """
        Return a list of all collection names in the current database.
        """
        return self.connection.list_collection_names()

    def drop_table(self, collection_name: str):
        """
        Drop the collection with the specified name.

        Args:
            collection_name: The name of the collection.

        Returns:
            The result of the drop operation.
        """
        return self.connection.drop_collection(collection_name)

    def insert(self, data):
        """
        Insert data into the current collection.
        Supports either a single dictionary or a list of dictionaries.

        Args:
            data: A dictionary or a list of dictionaries.

        Returns:
            If a single document is inserted, returns its inserted _id;
            if multiple documents are inserted, returns a list of inserted _id's.
        """
        assert self.collection is not None, "Collection not loaded. Use using() first"

        if isinstance(data, list):
            result = self.collection.insert_many(data)
            return result.inserted_ids
        elif isinstance(data, dict):
            result = self.collection.insert_one(data)
            return result.inserted_id
        else:
            raise ValueError("Inserted data must be a dict or a list of dicts")

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.collection = None
        return False

    @staticmethod
    def _build_query_mongo(mongo_query: dict) -> dict:
        """
        For MongoDB, the query condition already adheres to MongoDB's query syntax.
        Therefore, this method returns the original query (or an empty dict if None).

        Args:
            mongo_query: A dictionary representing the MongoDB query.

        Returns:
            The query dictionary.
        """
        if mongo_query is None or not mongo_query:
            return {}
        if not isinstance(mongo_query, dict):
            raise ValueError("Mongo query must be a dictionary")
        return mongo_query

    def select(self, query: dict = None):
        """
        Query the collection using MongoDB query syntax.

        Example:
            query = {"age": {"$gt": 30}, "$or": [{"name": "Alice"}, {"name": "Bob"}]}
            results = db.select(query)

        Returns:
            A list of documents matching the query.
        """
        assert self.collection is not None, "Collection not loaded. Use using() first"

        filter_query = self._build_query_mongo(query)
        # find() returns a Cursor; convert it to a list to retrieve all results.
        return list(self.collection.find(filter_query))

    def update(self, update_data: dict, query: dict = None):
        """
        Update documents in the collection that match the MongoDB query condition.

        Args:
            update_data: A dictionary specifying the fields and values to update.
                         (Internally wrapped with the $set operator.)
            query: A dictionary representing the MongoDB query.
                   If query is None or empty, all documents will be updated.

        Returns:
            The number of documents that were modified.
        """
        assert self.collection is not None, "Collection not loaded. Use using() first"

        filter_query = self._build_query_mongo(query)
        # Use the $set operator to update the fields.
        result = self.collection.update_many(filter_query, {"$set": update_data})
        return result.modified_count

    def delete(self, query: dict = None):
        """
        Delete documents in the collection that match the MongoDB query condition.

        Args:
            query: A dictionary representing the MongoDB query.
                   If query is None or empty, all documents will be deleted.

        Returns:
            The number of documents that were deleted.
        """
        assert self.collection is not None, "Collection not loaded. Use using() first"

        filter_query = self._build_query_mongo(query)
        result = self.collection.delete_many(filter_query)
        return result.deleted_count

    def close(self):
        """Close the MongoDB client connection."""
        self.connection.close()


if __name__ == "__main__":
    mongo_db = MongoDBDataBase()

    mongo_db.create_table("aarag")
    mongo_db.drop_table("aarag")

    mongo_db.create_table("fairy_tale")
    print(mongo_db.table_list())



from fastapi import APIRouter, HTTPException

from aa_rag.engine.lightrag import (
    LightRAGEngine,
    LightRAGInitParams,
    LightRAGRetrieveParams,
)
from aa_rag.engine.simple_chunk import (
    SimpleChunk,
    SimpleChunkRetrieveParams,
    SimpleChunkInitParams,
)
from aa_rag.gtypes.enums import EngineType
from aa_rag.gtypes.models.retrieve import (
    RetrieveItem,
    RetrieveResponse,
    SimpleChunkRetrieveItem,
    LightRAGRetrieveItem,
)

router = APIRouter(
    prefix="/retrieve", tags=["Retrieve"], responses={404: {"description": "Not found"}}
)


@router.post("/")
async def root(item: RetrieveItem):
    match item.engine_type:
        case EngineType.SimpleChunk:
            chunk_item = SimpleChunkRetrieveItem(**item.model_dump())
            return await chunk_retrieve(chunk_item)
        case _:
            raise HTTPException(status_code=400, detail="RetrieveType not supported")


@router.post("/chunk", tags=["SimpleChunk"])
async def chunk_retrieve(item: SimpleChunkRetrieveItem) -> RetrieveResponse:
    engine = SimpleChunk(SimpleChunkInitParams(**item.model_dump()))

    result = engine.retrieve(SimpleChunkRetrieveParams(**item.model_dump()))

    return RetrieveResponse(
        code=200,
        status="success",
        message=f"Retrieval completed via HybridRetrieve in {item.retrieve_mode}",
        data=RetrieveResponse.Data(documents=result),
    )


@router.post("/lightrag", tags=["LightRAG"])
async def lightrag_retrieve(item: LightRAGRetrieveItem) -> RetrieveResponse:
    engine = LightRAGEngine(LightRAGInitParams(**item.model_dump()))

    result = await engine.retrieve(LightRAGRetrieveParams(**item.model_dump()))

    return RetrieveResponse(
        code=200,
        status="success",
        message=f"Retrieval completed via LightRAGRetrieve in {item.retrieve_mode}",
        data=RetrieveResponse.Data(documents=result),
    )



import ast
import importlib
from pathlib import Path
from typing import Optional, Any, Literal

from dotenv.main import DotEnv
from pydantic import BaseModel, Field, SecretStr, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from aa_rag.gtypes.enums import (
    DBMode,
    RetrieveType,
    VectorDBType,
    NoSQLDBType,
    EngineType,
    LightRAGVectorStorageType,
    LightRAGGraphStorageType,
)


def load_env(key: str, default: Any = None):
    """
    Load environment variable from .env file. Convert to python object if possible.

    Args:
        key (str): Environment variable key.
        default (Any, optional): Default value if key not found. Defaults to None.If default is a tuple,
         it will be treated as a pair of values for development and production environments.

    Returns:
        Any: Python object representing the environment variable value or the default value.
    """
    env = DotEnv(Path(".env").absolute(), verbose=False, encoding="utf-8")

    if isinstance(default, tuple):
        env_mode = env.get("ENVIRONMENT")
        env_mode = env_mode.lower() if env_mode else "development"
        dev_dft, prod_dft = default
        match env_mode.lower():
            case "development" | "dev":
                default = dev_dft
            case "production" | "prod":
                default = prod_dft
            case _:
                default = dev_dft
    else:
        default = default

    value = env.get(key)

    if value is None:
        return default
    else:
        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            return value


class Server(BaseModel):
    host: str = Field(default="0.0.0.0", description="The host address for the server.")
    port: int = Field(
        default=222, description="The port number on which the server listens."
    )
    environment: Literal["Development", "Production"] = Field(
        default=load_env("ENVIRONMENT", ("Development", "Production")),
        description="The environment in which the server is running.",
    )


class OpenAI(BaseModel):
    api_key: SecretStr = Field(
        default=load_env("OPENAI_API_KEY"),
        alias="OPENAI_API_KEY",
        description="API key for accessing OpenAI services.",
        validate_default=True,
    )
    base_url: str = Field(
        default=load_env("OPENAI_BASE_URL", "https://api.openai.com/v1"),
        alias="OPENAI_BASE_URL",
        description="Base URL for OpenAI API requests.",
    )

    @field_validator("api_key")
    def check_api_key(cls, v):
        assert v.get_secret_value(), "API key is required."
        return v


class Storage(BaseModel):
    class LanceDB(BaseModel):
        uri: str = Field(
            default="./storage/lancedb",
            description="URI for lanceDB database location.",
        )

    class Milvus(BaseModel):
        uri: str = Field(
            default=load_env(
                "STORAGE_MILVUS_URI", ("./storage/milvus.db", "http://localhost:19530")
            ),
            description="URI for the Milvus server location.",
        )
        user: str = Field(default="", description="Username for the Milvus server.")
        password: SecretStr = Field(
            default="",
            description="Password for the Milvus server.",
            validate_default=True,
        )
        db_name: str = Field(
            default="aarag", description="Database name for the Milvus server."
        )

    class TinyDB(BaseModel):
        uri: str = Field(
            default="./storage/tinydb.json",
            description="URI for the relational database location.",
        )

    class MongoDB(BaseModel):
        uri: str = Field(
            default="mongodb://localhost:27017",
            description="URI for the MongoDB server location.",
        )
        user: str = Field(default="", description="Username for the MongoDB server.")
        password: SecretStr = Field(
            default="",
            description="Password for the MongoDB server.",
            validate_default=True,
        )
        db_name: str = Field(
            default="aarag", description="Database name for the MongoDB server."
        )

    class Neo4j(BaseModel):
        uri: str = Field(
            default=load_env("STORAGE_NEO4J_URI", (None, "bolt://localhost:7687")),
            description="URI for the Neo4j server location.",
        )
        user: str = Field(default=None, description="Username for the Neo4j server.")
        password: SecretStr = Field(
            default=None, description="Password for the Neo4j server."
        )

        @field_validator("uri")
        def check(cls, v):
            if v:
                if importlib.util.find_spec("neo4j") is None:
                    raise ImportError(
                        "Neo4j can only be enabled on the online service, please execute `pip install aa-rag[online]`."
                    )
            return v

    lancedb: LanceDB = Field(
        default_factory=LanceDB, description="LanceDB database configuration settings."
    )
    milvus: Milvus = Field(
        default_factory=Milvus, description="Milvus database configuration settings."
    )
    tinydb: TinyDB = Field(
        default_factory=TinyDB, description="TinyDB database configuration settings."
    )
    mongodb: MongoDB = Field(
        default_factory=MongoDB, description="MongoDB database configuration settings."
    )

    neo4j: Neo4j = Field(
        default_factory=Neo4j, description="Neo4j configuration settings."
    )

    mode: DBMode = Field(
        default=DBMode.UPSERT, description="Mode of operation for the database."
    )
    vector: VectorDBType = Field(
        default=VectorDBType.MILVUS, description="Type of vector database used."
    )
    nosql: NoSQLDBType = Field(
        default=load_env("DB_NOSQL", (NoSQLDBType.TINYDB, NoSQLDBType.MONGODB)),
        description="Type of NoSQL database used.",
    )

    @field_validator("vector")
    def check_vector_db(cls, v):
        if v == VectorDBType.LANCE:
            # check whether install lancedb package
            if importlib.util.find_spec("lancedb") is None:
                raise ImportError(
                    "LanceDB can only be enabled on the online service, please execute `pip install aa-rag[online]`."
                )
        return v

    @field_validator("nosql")
    def check_nosql_db(cls, v):
        if v == NoSQLDBType.MONGODB:
            # check whether install pymongo package
            if importlib.util.find_spec("pymongo") is None:
                raise ImportError(
                    "MongoDB can only be enabled on the online service, please execute `pip install aa-rag[online]`."
                )
        return v


class Embedding(BaseModel):
    model: str = Field(
        default="text-embedding-3-small",
        description="Model used for generating text embeddings.",
    )


class LLM(BaseModel):
    model: str = Field(
        default="gpt-4o",
        description="Model used for understanding text.",
    )
    multimodal_model: str = Field(
        default="gpt-4o",
        description="Model used for understanding the image.",
    )


class Engine(BaseModel):
    class SimpleChunk(BaseModel):
        class Index(BaseModel):
            chunk_size: int = Field(
                default=load_env("ENGINE_SIMPLECHUNK_INDEX_CHUNK_SIZE", 1000),
                description="Size of each chunk in the index.",
            )
            overlap_size: int = Field(
                default=load_env("ENGINE_SIMPLECHUNK_INDEX_OVERLAP_SIZE", 100),
                description="Overlap size between chunks in the index.",
            )

        class Retrieve(BaseModel):
            class Weight(BaseModel):
                dense: float = Field(
                    default=0.5, description="Weight for dense retrieval methods."
                )
                sparse: float = Field(
                    default=0.5, description="Weight for sparse retrieval methods."
                )

            k: int = Field(default=3, description="Number of top results to retrieve.")
            weight: Weight = Field(
                default_factory=Weight,
                description="Weights for different retrieval methods.",
            )
            type: RetrieveType = Field(
                default=RetrieveType.HYBRID,
                description="Type of retrieval strategy used.",
            )

        index: Index = Field(
            default_factory=Index, description="Index configuration settings."
        )
        retrieve: Retrieve = Field(
            default_factory=Retrieve, description="Retrieve configuration settings."
        )

    class LightRAG(BaseModel):
        dir: str = Field(
            default="./storage/lightrag",
            description="Directory for LightRAG database location.",
        )
        vector_storage: LightRAGVectorStorageType = Field(
            default=LightRAGVectorStorageType.MILVUS,
            description="Type of vector storage used for LightRAG.",
        )
        graph_storage: LightRAGGraphStorageType = Field(
            default=load_env(
                "LIGHTRAG_GRAPH_STORAGE",
                (LightRAGGraphStorageType.NETWORKX, LightRAGGraphStorageType.NEO4J),
            ),
            description="Type of graph storage used for LightRAG.",
        )
        llm: str = Field(
            default="gpt-4o-mini", description="Model used for understanding text."
        )
        embedding: str = Field(
            default_factory=Embedding,
            description="Model used for generating text embeddings.",
        )
        cosine_threshold: float = Field(
            default=load_env("LIGHTRAG_COSINE_THRESHOLD", 0.3),
            description="Cosine similarity threshold for LightRAG.",
        )

        k: int = Field(
            default=60,
            description="Number of top items to retrieve. Represents entities in 'local' mode and relationships in 'global' mode.",
        )

        @field_validator("graph_storage")
        def check(cls, v):
            if v == LightRAGGraphStorageType.NEO4J:
                if importlib.util.find_spec("neo4j") is None:
                    raise ImportError(
                        "Neo4j can only be enabled on the online service, please execute `pip install aa-rag[online]`."
                    )
            return v

    type: EngineType = Field(
        default=EngineType.SimpleChunk,
        description="Type of index used for data retrieval.",
    )

    simple_chunk: SimpleChunk = Field(
        default_factory=SimpleChunk,
        description="Simple chunk index configuration settings.",
    )

    lightrag: LightRAG = Field(
        default_factory=LightRAG,
        description="LightRAG index configuration settings.",
    )


class Retrieve(BaseModel):
    class Weight(BaseModel):
        dense: float = Field(
            default=0.5, description="Weight for dense retrieval methods."
        )
        sparse: float = Field(
            default=0.5, description="Weight for sparse retrieval methods."
        )

    type: RetrieveType = Field(
        default=RetrieveType.HYBRID, description="Type of retrieval strategy used."
    )
    k: int = Field(default=3, description="Number of top results to retrieve.")
    weight: Weight = Field(
        default_factory=Weight, description="Weights for different retrieval methods."
    )


class OSS(BaseModel):
    access_key: Optional[str] = Field(
        default=load_env("OSS_ACCESS_KEY"),
        alias="OSS_ACCESS_KEY",
        description="Access key for accessing OSS services.",
    )

    endpoint: str = Field(
        default="https://s3.amazonaws.com",
        description="Endpoint for OSS API requests.",
    )
    secret_key: Optional[SecretStr] = Field(
        default=load_env("OSS_SECRET_KEY"),
        alias="OSS_SECRET_KEY",
        description="Secret key for accessing OSS services.",
        validate_default=True,
    )

    bucket: str = Field(default="aarag", description="Bucket name for storing data.")
    cache_bucket: str = Field(
        default=load_env("OSS_CACHE_BUCKET", "aarag-cache"),
        description="Bucket name for storing cache data.",
    )

    @field_validator("access_key")
    def check_access_key(cls, v):
        if v:
            if importlib.util.find_spec("boto3") is None:
                raise ImportError(
                    "OSS can only be enabled on the online service, please execute `pip install aa-rag[online]`."
                )
        return v


class Settings(BaseSettings):
    server: Server = Field(
        default_factory=Server, description="Server configuration settings."
    )
    openai: OpenAI = Field(
        default_factory=OpenAI, description="OpenAI API configuration settings."
    )

    storage: Storage = Field(
        default_factory=Storage, description="Database configuration settings."
    )
    embedding: Embedding = Field(
        default_factory=Embedding, description="Embedding model configuration settings."
    )
    engine: Engine = Field(
        default_factory=Engine, description="Indexing engine configuration settings."
    )
    llm: LLM = Field(
        default_factory=LLM,
        description="Language model configuration settings.",
    )

    oss: OSS = Field(default_factory=OSS, description="Minio configuration settings.")

    # 这里禁用了自动的 CLI 解析
    model_config = SettingsConfigDict(
        env_file=".env",
        env_nested_delimiter="_",
        extra="ignore",
    )


setting = Settings()
